import sklearn

from . import _sklearn_model


class RandomForest(_sklearn_model.SklearnModel):
    """
    Implementation of a class for Random Forest.

    See :obj:`~easypheno.model._base_model.BaseModel` for more information on the attributes.
    """
    standard_encoding = '012'
    possible_encodings = ['012']

    def define_model(self):
        """
        Definition of the actual prediction model.

        See :obj:`~easypheno.model._base_model.BaseModel` for more information.
        """
        # all hyperparameters defined are suggested for optimization
        params = self.suggest_all_hyperparams_to_optuna()
        # add random_state for reproducibility and n_jobs for multiprocessing
        params.update({'random_state': 42, 'n_jobs': -1})
        if self.task == 'classification':
            return sklearn.ensemble.RandomForestClassifier(**params)
        else:
            return sklearn.ensemble.RandomForestRegressor(**params)

    def define_hyperparams_to_tune(self) -> dict:
        """
        See :obj:`~easypheno.model._base_model.BaseModel` for more information on the format.
        """
        return {
            'n_estimators': {
                'datatype': 'categorical',
                'list_of_values': [50, 100, 250, 500, 750, 1000, 1250, 1500, 1750, 2000, 2250, 2500, 2750, 3000,
                                   3500, 4000, 4500, 5000]
            },
            'min_samples_split': {
                'datatype': 'float',
                'lower_bound': 0.005,
                'upper_bound': 0.2,
                'step': 0.005
            },
            'max_depth': {
                'datatype': 'int',
                'lower_bound': 2,
                'upper_bound': 50,
                'step': 2
            },
            'min_samples_leaf': {
                'datatype': 'float',
                'lower_bound': 0.005,
                'upper_bound': 0.2,
                'step': 0.005
            },
            'max_features': {
                'datatype': 'categorical',
                'list_of_values': ['sqrt', 'log2']
            }
        }



import sklearn

from . import _sklearn_model


class SupportVectorMachine(_sklearn_model.SklearnModel):
    """
    Implementation of a class for Support Vector Machine respective Regression.

    See :obj:`~easypheno.model._base_model.BaseModel` for more information on the attributes.
    """
    standard_encoding = '012'
    possible_encodings = ['012']

    def define_model(self):
        """
        Definition of the actual prediction model.

        See :obj:`~easypheno.model._base_model.BaseModel` for more information.
        """
        kernel = self.suggest_hyperparam_to_optuna('kernel')
        reg_c = self.suggest_hyperparam_to_optuna('C')
        if kernel == 'poly':
            degree = self.suggest_hyperparam_to_optuna('degree')
            gamma = self.suggest_hyperparam_to_optuna('gamma')
        elif kernel in ['rbf', 'sigmoid']:
            degree = 42  # default
            gamma = self.suggest_hyperparam_to_optuna('gamma')
        elif kernel == 'linear':
            degree = 42  # default
            gamma = 42  # default
        if self.task == 'classification':
            return sklearn.svm.SVC(kernel=kernel, C=reg_c, degree=degree, gamma=gamma, random_state=42,
                                   max_iter=1000000)
        else:
            return sklearn.svm.SVR(kernel=kernel, C=reg_c, degree=degree, gamma=gamma, max_iter=1000000)

    def define_hyperparams_to_tune(self) -> dict:
        """
        See :obj:`~easypheno.model._base_model.BaseModel` for more information on the format.
        """
        return {
            'kernel': {
                'datatype': 'categorical',
                'list_of_values': ['linear', 'poly', 'rbf'],
            },
            'degree': {
                'datatype': 'int',
                'lower_bound': 1,
                'upper_bound': 5
            },
            'gamma': {
                'datatype': 'float',
                'lower_bound': 10**-3,
                'upper_bound': 10**3,
                'log': True
            },
            'C': {
                'datatype': 'float',
                'lower_bound': 10**-3,
                'upper_bound': 10**3,
                'log': True
            }
        }



import argparse

from easypheno.utils import helper_functions
from easypheno.preprocess import encoding_functions
from . import optim_pipeline

if __name__ == '__main__':
    """
    Run file to start the whole procedure:
            Parameter Plausibility Check
            Check and prepare data files
            Bayesian optimization for each chosen model
    """
    # User Input #
    parser = argparse.ArgumentParser()
    # Input Params #
    parser.add_argument("-dd", "--data_dir", type=str,
                        default='/myhome/easyPheno/docs/source/tutorials/tutorial_data/',
                        help="Provide the full path of your data directory (that contains the geno- and phenotype "
                             "files).")
    parser.add_argument("-sd", "--save_dir", type=str, default='/myhome/',
                        help="Provide the full path of the directory in which you want to save your results. "
                             "Default is same as data_dir")
    parser.add_argument("-gm", "--genotype_matrix", type=str, default='x_matrix.csv',
                        help="specify the name (including data type suffix) of the genotype matrix to be used. "
                             "Needs to be located in the specified data_dir."
                             "For more info regarding the required format see our documentation.")
    parser.add_argument("-pm", "--phenotype_matrix", type=str, default='y_matrix.csv',
                        help="specify the name (including data type suffix) of the phenotype matrix to be used. "
                              "Needs to be located in the specified data_dir."
                             "For more info regarding the required format see our documentation.")
    parser.add_argument("-ph", "--phenotype", nargs='+', type=str, default=['continuous_values'],
                        help="specify the name of the phenotype to be predicted. "
                             "Multiple phenotypes can also be chosen if they are in the same phenotype matrix. "
                             "Just name the phenotypes, e.g. --phenotype FT10 FT16")
    parser.add_argument("-enc", "--encoding", type=str, default=None,
                        help="specify the encoding to use. Caution: has to be a possible encoding for the model to use."
                             "Valid arguments are: " + str(encoding_functions.get_list_of_encodings()))

    # Preprocess Params #
    parser.add_argument("-maf", "--maf_percentage", type=int, default=0,
                        help="specify the minor allele frequency (as percentage value). "
                             "specify 0 if you do not want a maf filter.")
    parser.add_argument("-split", "--datasplit", type=str, default='nested-cv',
                        help="specify the data split to use: 'nested-cv' | 'cv-test' | 'train-val-test'"
                             "Default values are 5 folds, train-test-split to 80/20 and train-val-test to 60/20/20")
    parser.add_argument("-testperc", "--test_set_size_percentage", type=int, default=20,
                        help="specify the size of the test set in percentage. "
                             "Standard is 20, only relevant for 'cv-test' and 'train-val-test'")
    parser.add_argument("-valperc", "--val_set_size_percentage", type=int, default=20,
                        help="specify the size of the validation set in percentage. "
                             "Standard is 20, only relevant for 'train-val-test'")
    parser.add_argument("-of", "--n_outerfolds", type=int, default=5,
                        help="specify the number of outerfolds to use for 'nested_cv'"
                             "Standard is 5, only relevant for 'nested_cv'")
    parser.add_argument("-folds", "--n_innerfolds", type=int, default=5,
                        help="specify the number of innerfolds/folds to use for 'nested_cv' respectively 'cv-test'"
                             "Standard is 5, only relevant for 'nested_cv' and 'cv-test'")

    # Model and Optimization Params #
    parser.add_argument("-mod", "--models", nargs='+', type=str, default=['xgboost'],
                        help="specify the models to optimize: 'all' or naming according to source file name. "
                             "Multiple models can be selected by just naming multiple model names, "
                             "e.g. --models mlp xgboost. "
                             "The following are available: " + str(helper_functions.get_list_of_implemented_models()))
    parser.add_argument("-tr", "--n_trials", type=int, default=10,
                        help="number of trials for optuna")
    parser.add_argument("-sf", "--save_final_model", type=lambda x: (str(x).lower() in ['true', '1', 'yes']),
                        default=False,
                        help="save the final model to hard drive "
                             "(caution: some models may use a lot of disk space, "
                             "unfitted models that can be retrained are already saved by default)")

    # Only relevant for Neural Networks #
    parser.add_argument("-bs", "--batch_size", type=int, default=None,
                        help="Only relevant for neural networks: define the batch size.")
    parser.add_argument("-ep", "--n_epochs", type=int, default=None,
                        help="Only relevant for neural networks: define the number of epochs. If nothing is specified,"
                             "it will be considered as a hyperparameter for optimization")

    parser.add_argument("-ofn", "--outerfold_number_to_run", type=int, default=None,
                        help="Use this parameter in case you only want to run the optimization for one outer fold, "
                             "counting starts at 0")

    args = vars(parser.parse_args())
    phenotypes = args["phenotype"]

    for phenotype in phenotypes:
        args["phenotype"] = phenotype
        try:
            optim_pipeline.run(**args)
        except Exception as exc:
            print("Failure when running pipeline for " + phenotype)
            print(exc)
            continue



import argparse
import numpy as np
import pandas as pd
import pathlib
import shutil

from ..model import _model_functions
from ..utils import check_functions, helper_functions
from ..preprocess import base_dataset
from ..evaluation import eval_metrics


if __name__ == "__main__":
    """
    Rerun inference using best hyperparameters on already-used data
    """
    parser = argparse.ArgumentParser()
    parser.add_argument("-dd", "--data_dir", type=str, default=None,
                        help="Provide the full path of the old data directory (that contains the geno- and phenotype "
                             "files as well as the index file the model was trained on). "
                             "Only needed if the final model was not saved.")
    parser.add_argument("-gm", "--genotype_matrix", type=str,
                        help="Provide the name of the new genotype matrix you want to predict on")
    parser.add_argument("-pm", "--phenotype_matrix", type=str,
                        help="Provide the name of the new phenotype matrix you want to predict on")
    parser.add_argument("-sd", "--save_dir", type=str,
                        help="Define the save directory for the results.")
    parser.add_argument("-rd", "--results_dir_model", type=str,
                        help="Provide the full path of the directory where your results of the model "
                             "you want to use are stored")
    args = vars(parser.parse_args())
    data_dir = args['data_dir']
    genotype_matrix = args['genotype_matrix']
    phenotype_matrix = args['phenotype_matrix']
    save_dir = args["save_dir"]
    results_directory_model = args['results_dir_model']
    str_results_dir_model = args['results_dir_model']
    overall_results = {}
    for fold in range(3):
        helper_functions.set_all_seeds(42)
        results_directory_model = \
            pathlib.Path(str_results_dir_model[:str_results_dir_model.rindex('outerfold')] +
                         'outerfold_' + str(fold) +
                         str_results_dir_model[str_results_dir_model.rindex('outerfold') + 11:])

        data_dir = pathlib.Path(data_dir)
        save_dir = pathlib.Path(save_dir)
        new_results_dir = save_dir.joinpath(*results_directory_model.parts[-6:])
        check_functions.check_exist_directories(list_of_dirs=[new_results_dir], create_if_not_exist=True)
        for file in helper_functions.get_all_files(results_directory_model):
            shutil.copyfile(file, new_results_dir.joinpath(file.name))
        shutil.move(new_results_dir.joinpath('final_model_test_results.csv'),
                    new_results_dir.joinpath('final_model_test_results_old.csv'))
        # Check user inputs
        print("Checking user inputs")
        if not check_functions.check_exist_directories(
                list_of_dirs=[results_directory_model, data_dir, save_dir]):
            raise Exception("See output above. Problems with specified directories")

        result_folder_name = results_directory_model.parts[-3] if 'nested' in str(results_directory_model) \
            else results_directory_model.parts[-2]
        model_name = results_directory_model.parts[-1]
        if 'fromR' in model_name:
            import easypheno.model
            easypheno.model.__all__.extend(['_bayesfromR', 'bayesAfromR', 'bayesBfromR', 'bayesCfromR'])
        datasplit, n_outerfolds, n_innerfolds, val_set_size_percentage, test_set_size_percentage, maf_perc = \
            helper_functions.get_datasplit_config_info_for_resultfolder(resultfolder=result_folder_name)
        nested_offset = -1 if 'nested' in datasplit else 0
        phenotype = results_directory_model.parts[-3 + nested_offset]
        if not check_functions.check_exist_files(
                list_of_files=[
                    data_dir.joinpath(phenotype_matrix), data_dir.joinpath(genotype_matrix),
                    data_dir.joinpath(base_dataset.Dataset.get_index_file_name(
                        genotype_matrix_name=genotype_matrix, phenotype_matrix_name=phenotype_matrix,
                        phenotype=phenotype))]):
            raise Exception("See output above. Problems with specified files.")


        encoding = helper_functions.get_mapping_name_to_class()[model_name].standard_encoding

        # Prepare the model
        outerfold_number = int(results_directory_model.parent.parts[-1].split('_')[1]) if 'nested' in datasplit else 0
        models = results_directory_model.parts[-2 + nested_offset].split('_')[3].split('+')
        results_file_path = \
            results_directory_model.parents[0 - nested_offset].joinpath('Results_overview_' + '_'.join(models) + '.csv')
        print("Retraining model")
        print("Loading old dataset")
        dataset = base_dataset.Dataset(
            data_dir=data_dir, genotype_matrix_name=genotype_matrix,
            phenotype_matrix_name=phenotype_matrix,
            phenotype=phenotype, datasplit=datasplit, n_outerfolds=n_outerfolds, n_innerfolds=n_innerfolds,
            test_set_size_percentage=test_set_size_percentage, val_set_size_percentage=val_set_size_percentage,
            encoding=encoding, maf_percentage=maf_perc
        )
        if not results_file_path.is_file():
            raise Exception("Results Overview file not existing. Please check: " + str(results_file_path))
        model = _model_functions.retrain_model_with_results_file(
            results_file_path=results_file_path, model_name=model_name, datasplit=datasplit,
            outerfold_number=outerfold_number, dataset=dataset
        )

        # Do inference and save results
        print('-----------------------------------------------')
        print("Inference on new data for " + model_name)
        outerfold_info = dataset.datasplit_indices['outerfold_' + str(outerfold_number)]
        X_test, y_test, sample_ids_test = \
            dataset.X_full[outerfold_info['test']], dataset.y_full[outerfold_info['test']], \
            dataset.sample_ids_full[outerfold_info['test']]
        X_retrain, y_retrain, sample_ids_retrain = \
            dataset.X_full[~np.isin(np.arange(len(dataset.X_full)), outerfold_info['test'])], \
            dataset.y_full[~np.isin(np.arange(len(dataset.y_full)), outerfold_info['test'])], \
            dataset.sample_ids_full[~np.isin(np.arange(len(dataset.sample_ids_full)),
                                                  outerfold_info['test'])]
        y_pred_retrain = model.predict(X_in=X_retrain)
        y_pred_test = model.predict(X_in=X_test)
        eval_scores = \
            eval_metrics.get_evaluation_report(y_pred=y_pred_test, y_true=y_test, task=model.task,
                                               prefix='test_')
        print('## Results on test set ##')
        print(eval_scores)
        final_results = pd.DataFrame(index=range(0, dataset.y_full.shape[0]))
        final_results.at[0:len(sample_ids_retrain) - 1, 'sample_ids_retrain'] = sample_ids_retrain.flatten()
        final_results.at[0:len(y_pred_retrain) - 1, 'y_pred_retrain'] = y_pred_retrain.flatten()
        final_results.at[0:len(y_retrain) - 1, 'y_true_retrain'] = y_retrain.flatten()
        final_results.at[0:len(sample_ids_test) - 1, 'sample_ids_test'] = sample_ids_test.flatten()
        final_results.at[0:len(y_pred_test) - 1, 'y_pred_test'] = y_pred_test.flatten()
        final_results.at[0:len(y_test) - 1, 'y_true_test'] = y_test.flatten()
        for metric, value in eval_scores.items():
            final_results.at[0, metric] = value
        final_results.to_csv(new_results_dir.joinpath(
            'final_model_test_results.csv'),
            sep=',', decimal='.', float_format='%.10f', index=False
        )
        key = 'outerfold_' + str(fold)
        best_params = model.optuna_trial.params
        runtime_metrics = {
            'process_time_mean': 0, 'process_time_std': 0,
            'process_time_max': 0, 'process_time_min': 0,
            'real_time_mean': 0, 'real_time_std': 0,
            'real_time_max': 0, 'real_time_min': 0
        }
        overall_results[key] = {'best_params': best_params, 'eval_metrics': eval_scores,
                                'runtime_metrics': runtime_metrics}
    model_overview = {model_name: overall_results}
    path_overview_file = new_results_dir.parent.parent.joinpath(
        'Results_overview_' + '_'.join([model_name]) + '.csv')
    helper_functions.save_model_overview_dict(model_overview=model_overview, save_path=path_overview_file)



import sklearn
import numpy as np


def get_evaluation_report(y_pred: np.array, y_true: np.array, task: str, prefix: str = '') -> dict:
    """
    Get values for common evaluation metrics

    :param y_pred: predicted values
    :param y_true: true values
    :param task: ML task to solve
    :param prefix: prefix to be added to the key if multiple eval metrics are collected

    :return: dictionary with common metrics
    """
    if len(y_pred) == (len(y_true)-1):
        print('y_pred has one element less than y_true (e.g. due to batch size config) -> dropped last element')
        y_true = y_true[:-1]
    if task == 'classification':
        average = 'micro' if len(np.unique(y_true)) > 2 else 'binary'
        eval_report_dict = {
            prefix + 'accuracy': sklearn.metrics.accuracy_score(y_true=y_true, y_pred=y_pred),
            prefix + 'f1_score': sklearn.metrics.f1_score(y_true=y_true, y_pred=y_pred, average=average),
            prefix + 'precision': sklearn.metrics.precision_score(y_true=y_true, y_pred=y_pred,
                                                                  zero_division=0, average=average),
            prefix + 'recall': sklearn.metrics.recall_score(y_true=y_true, y_pred=y_pred,
                                                            zero_division=0, average=average),
            prefix + 'mcc': sklearn.metrics.matthews_corrcoef(y_true=y_true, y_pred=y_pred)
        }
    else:
        eval_report_dict = {
            prefix + 'mse': sklearn.metrics.mean_squared_error(y_true=y_true, y_pred=y_pred),
            prefix + 'rmse': sklearn.metrics.mean_squared_error(y_true=y_true, y_pred=y_pred, squared=False),
            prefix + 'r2_score': sklearn.metrics.r2_score(y_true=y_true, y_pred=y_pred),
            prefix + 'explained_variance': sklearn.metrics.explained_variance_score(y_true=y_true, y_pred=y_pred)
        }
    return eval_report_dict



import argparse
import pathlib

from ..preprocess import raw_data_functions, encoding_functions
from ..utils import check_functions
from .synthetic_phenotypes import save_simulation


if __name__ == "__main__":
    """
    Run file to generate synthetic phenotypes
    """

    parser = argparse.ArgumentParser()
    parser.add_argument("-dd", "--data_dir", type=str,
                        help="Provide the full path of the directory where your genotype data is stored")
    parser.add_argument("-gm", "--genotype_matrix", type=str,
                        help="Specify the name (including data type suffix) of the genotype matrix to be used. "
                             "Needs to be located in the specified data_dir."
                             "For more info regarding the required format see our documentation.")
    parser.add_argument("-nsim", "--number_of_simulations", type=int, default=1,
                        help="Specify the number of simulations to create with the same configurations.")
    parser.add_argument("-nsamp", "--number_of_samples", type=int, default=1000,
                        help="Specify the number of samples the simulated phenotype should have. The maximum is the "
                             "number of samples of the genotype matrix")
    parser.add_argument("-ncaus", "--number_causal_snps", type=int, default=1,
                        help="Specify the number of causal SNPs for the simulation. Need at least one causal SNP.")
    parser.add_argument("-nback", "--number_background_snps", type=int, default=1000,
                        help="Specify the number of SNPs used to simulate the polygenic background. Need at least one "
                             "SNP.")
    parser.add_argument("-ev", "--explained_variance", type=int, default=30,
                        help="Specify the total explained variance of the causal SNPs, i.e. how much of the total "
                             "variance of the phenotype should be explained by the causal SNPs.")
    parser.add_argument("-maf", type=int, default=0,
                        help="Specify the minor allele frequency (as percentage value). "
                             "specify 0 if you do not want a maf filter.")
    parser.add_argument("-her", "--heritability", type=int, default=70,
                        help="Specify the heritability of the phenotype, i.e. how much of the variance of "
                             "polygenic background + noise should be explained by the polygenic background.")
    parser.add_argument("-seed", type=int, default=42,
                        help="Specify the seed for random sampling.")
    parser.add_argument("-dist", "--distribution", type=str, default="normal",
                        help="Specify the distribution of the noise. Can be 'gamma' or 'normal'")
    parser.add_argument("-shape", type=float, default=None,
                        help="If distribution is 'gamma', specify the shape parameter of the distribution. If not "
                             "specified it will be set to 1.0")
    parser.add_argument("-sd", "--save_dir", type=str, default=None,
                        help="Define save directory for the synthetic phenotypes. Default is the data directory.")
    args = vars(parser.parse_args())
    data_dir = pathlib.Path(args['data_dir'])
    if args['save_dir'] is None:
        args['save_dir'] = args['data_dir']
    if args['distribution'] == 'gamma' and args['shape'] is None:
        args['shape'] = 1.0

    # load genotype
    check_functions.check_exist_directories(list_of_dirs=[data_dir], create_if_not_exist=False)
    check_functions.check_exist_files([data_dir.joinpath(args['genotype_matrix'])])
    X, sample_ids, snp_ids = raw_data_functions.check_transform_format_genotype_matrix(data_dir=data_dir,
                            genotype_matrix_name=args['genotype_matrix'], models=None, user_encoding='012')
    X = encoding_functions.get_additive_encoding(X)
    print('Have genotype matrix ', args['genotype_matrix'])

    # create and save simulations
    save_simulation(save_dir=args['save_dir'], genotype_matrix_name=args['genotype_matrix'],
                    number_of_sim=args['number_of_simulations'], X=X, sample_ids=sample_ids,
                    snp_ids=snp_ids, number_of_samples=args['number_of_samples'],
                    number_causal_snps=args['number_causal_snps'], explained_variance=args['explained_variance'],
                    maf=args['maf'], heritability=args['heritability'], seed=args['seed'],
                    number_background_snps=args['number_background_snps'], distribution=args['distribution'],
                    shape=args['shape'])



import argparse

from . import feat_importance

if __name__ == "__main__":
    """
    Run to generate feature importance .csv files based on result files of already finished optimizations
    """
    parser = argparse.ArgumentParser()
    parser.add_argument("-dd", "--data_dir", type=str,
                        help="Provide the full path of your data directory (that contains the geno- and phenotype "
                             "files as well as the index file).")
    parser.add_argument("-rd", "--results_dir", type=str,
                        help="Provide the full path of the directory where your results are stored and for which "
                             "you want to post-generate feature importances")
    args = vars(parser.parse_args())
    data_dir = args['data_dir']
    results_directory_genotype_level = args['results_dir']

    feat_importance.post_generate_feature_importances(
        results_directory_genotype_level=results_directory_genotype_level, data_dir=data_dir
    )



from . import _bayesfromR


class BayesB(_bayesfromR.Bayes_R):
    """
    Implementation of a class for Bayes B.

    *Attributes*

        *Inherited attributes*

        See :obj:`~easypheno.model._bayesfromR.Bayes_R` for more information on the attributes.
    """

    def __init__(self, task: str, encoding: str = None):
        super().__init__(task=task, model_name='BayesB', encoding=encoding)



import warnings
import os
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'

with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    import easypheno.evaluation as evaluation
    import easypheno.model as model
    import easypheno.utils as utils
    import easypheno.optimization as optimization
    import easypheno.preprocess as preprocess
    import easypheno.postprocess as postprocess
    import easypheno.simulate as simulate

    from . import optim_pipeline

__version__ = "0.0.5"
__author__ = 'Florian Haselbeck, Maura John, Dominik G. Grimm'
__credits__ = 'GrimmLab @ TUM Campus Straubing (https://bit.cs.tum.de/)'



import sklearn

from . import _sklearn_model


class ElasticNet(_sklearn_model.SklearnModel):
    """
    Implementation of a class for Linear respective Logistic Regression using ElasticNet penalty.

    See :obj:`~easypheno.model._base_model.BaseModel` for more information on the attributes.
    """
    standard_encoding = '012'
    possible_encodings = ['012']

    def define_model(self):
        """
        Definition of the actual prediction model.

        See :obj:`~easypheno.model._base_model.BaseModel` for more information.
        """
        # Penalty term is fixed to l1, but might also be optimized
        penalty = 'elasticnet'  # self.suggest_hyperparam_to_optuna('penalty')
        if penalty == 'l1':
            l1_ratio = 1
        elif penalty == 'l2':
            l1_ratio = 0
        else:
            l1_ratio = self.suggest_hyperparam_to_optuna('l1_ratio')
        if self.task == 'classification':
            reg_c = self.suggest_hyperparam_to_optuna('C')
            return sklearn.linear_model.LogisticRegression(penalty=penalty, C=reg_c, solver='saga',
                                                           l1_ratio=l1_ratio if penalty == 'elasticnet' else None,
                                                           max_iter=10000, random_state=42, n_jobs=-1)
        else:
            alpha = self.suggest_hyperparam_to_optuna('alpha')
            return sklearn.linear_model.ElasticNet(alpha=alpha, l1_ratio=l1_ratio, max_iter=10000, random_state=42)

    def define_hyperparams_to_tune(self) -> dict:
        """
        See :obj:`~easypheno.model._base_model.BaseModel` for more information on the format.
        """
        return {
            'penalty': {
                'datatype': 'categorical',
                'list_of_values': ['l1', 'l2', 'elasticnet']
            },
            'l1_ratio': {
                'datatype': 'float',
                'lower_bound': 0.05,
                'upper_bound': 0.95,
                'step': 0.05
            },
            'alpha': {
                'datatype': 'float',
                'lower_bound': 10**-3,
                'upper_bound': 10**3,
                'log': True
            },
            'C': {
                'datatype': 'float',
                'lower_bound': 10**-3,
                'upper_bound': 10**3,
                'log': True
            }
        }



import os
import inspect
import importlib
import torch
import random
import numpy as np
import pandas as pd
import tensorflow as tf
import pathlib

import easypheno.model


def get_list_of_implemented_models() -> list:
    """
    Create a list of all implemented models based on files existing in 'model' subdirectory of the repository.
    """
    # Assumption: naming of python source file is the same as the model name specified by the user
    if os.path.exists('../model'):
        model_src_files = os.listdir('../model')
    elif os.path.exists('model'):
        model_src_files = os.listdir('model')
    elif os.path.exists('easypheno/model'):
        model_src_files = os.listdir('easypheno/model')
    else:
        model_src_files = [model_file + '.py' for model_file in easypheno.model.__all__]
    model_src_files = [file for file in model_src_files if file[0] != '_']
    return [model[:-3] for model in model_src_files]


def test_likely_categorical(vector_to_test: list, abs_unique_threshold: int = 20) -> bool:
    """
    Test whether a vector is most likely categorical.
    Simple heuristics: checking if the number of unique values exceeds a specified threshold

    :param vector_to_test: vector that is tested if it is most likely categorical
    :param abs_unique_threshold: threshold of unique values' ratio to declare vector categorical

    :return: True if the vector is most likely categorical, False otherwise
    """
    number_unique_values = np.unique(vector_to_test).shape[0]
    return number_unique_values <= abs_unique_threshold


def get_mapping_name_to_class() -> dict:
    """
    Get a mapping from model name (naming in package model without .py) to class name.

    :return: dictionary with mapping model name to class name
    """
    if os.path.exists('../model'):
        files = os.listdir('../model')
    elif os.path.exists('model'):
        files = os.listdir('model')
    elif os.path.exists('easypheno/model'):
        files = os.listdir('easypheno/model')
    else:
        files = [model_file + '.py' for model_file in easypheno.model.__all__]
    modules_mapped = {}
    for file in files:
        if file not in ['__init__.py', '__pycache__']:
            if file[-3:] != '.py':
                continue

            file_name = file[:-3]
            module_name = 'easypheno.model.' + file_name
            for name, cls in inspect.getmembers(importlib.import_module(module_name), inspect.isclass):
                if cls.__module__ == module_name:
                    modules_mapped[file_name] = cls
    return modules_mapped


def set_all_seeds(seed: int = 42):
    """
    Set all seeds of libs with a specific function for reproducibility of results

    :param seed: seed to use
    """
    torch.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)
    torch.cuda.manual_seed(seed)
    random.seed(seed)
    np.random.seed(seed)
    torch.backends.cudnn.deterministic = True
    torch.backends.cudnn.benchmark = False
    tf.random.set_seed(seed)


def get_subpath_for_datasplit(datasplit: str, datasplit_params: list) -> str:
    """
    Construct the subpath according to the datasplit.

    Datasplit parameters:

    - nested-cv: [n_outerfolds, n_innerfolds]
    - cv-test: [n_innerfolds, test_set_size_percentage]
    - train-val-test: [val_set_size_percentage, test_set_size_percentage]

    :param datasplit: datasplit to retrieve
    :param datasplit_params: parameters to use for the specific datasplit

    :return: string with the subpath
    """
    # construct subpath due to the specified datasplit
    if datasplit == 'train-val-test':
        datasplit_string = f'({100 - datasplit_params[0]}-{datasplit_params[0]})-{datasplit_params[1]}'
    elif datasplit == 'cv-test':
        datasplit_string = f'{datasplit_params[0]}-{datasplit_params[1]}'
    elif datasplit == 'nested-cv':
        datasplit_string = f'{datasplit_params[0]}-{datasplit_params[1]}'
    return datasplit_string


def save_model_overview_dict(model_overview: dict, save_path: str):
    """
    Structure and save results of a whole optimization run for multiple models in one csv file

    :param model_overview: dictionary with results overview
    :param save_path: filepath for saving the results overview file
    """
    results_overview = pd.DataFrame()
    for model_name, fold_dicts in model_overview.items():
        result_dicts = {}
        result_dicts_std = {}
        runtime_dicts = {}
        runtime_dicts_std = {}
        for fold_name, fold_info in fold_dicts.items():
            for result_name, result_info in fold_info.items():
                results_overview.at[fold_name, model_name + '___' + result_name] = [result_info]
                if 'eval_metric' in result_name:
                    for metric_name, metric_result in result_info.items():
                        if metric_name not in result_dicts.keys():
                            result_dicts[metric_name] = []
                        result_dicts[metric_name].append(metric_result)
                if 'runtime' in result_name:
                    for metric_name, metric_result in result_info.items():
                        if metric_name not in runtime_dicts.keys():
                            runtime_dicts[metric_name] = []
                        runtime_dicts[metric_name].append(metric_result)
        for metric_name, results in result_dicts.items():
            result_dicts[metric_name] = np.mean(results)
            result_dicts_std[metric_name] = np.std(results)
        for metric_name, results in runtime_dicts.items():
            runtime_dicts[metric_name] = np.mean(results)
            runtime_dicts_std[metric_name] = np.std(results)
        if 'nested' in str(save_path):
            results_overview.at['mean_over_all_folds', model_name + '___' + 'eval_metrics'] = [result_dicts]
            results_overview.at['std_over_all_folds', model_name + '___' + 'eval_metrics'] = [result_dicts_std]
            results_overview.at['mean_over_all_folds', model_name + '___' + 'runtime_metrics'] = [runtime_dicts]
            results_overview.at['std_over_all_folds', model_name + '___' + 'runtime_metrics'] = [runtime_dicts_std]
    results_overview.to_csv(save_path)


def sort_models_by_encoding(models_list: list) -> list:
    """
    Sort models by the encoding that will be used

    :param models_list: unsorted list of models

    :return: list of models sorted by encoding
    """
    encodings = [get_mapping_name_to_class()[model_name].standard_encoding for model_name in models_list]
    sorted_models_list = [el[0] for el in sorted(zip(models_list, encodings), key=lambda x: x[1])]
    return sorted_models_list


def get_all_subdirectories_non_recursive(path: pathlib.Path) -> list:
    """
    Get all non-recursive subdirectories of path

    :param path: path to search

    :return: list with all non-recursive subdirs
    """
    return list(filter(lambda x: x.is_dir(), path.iterdir()))


def get_all_files(path: pathlib.Path) -> list:
    """
    Get all non-recursive files of path

    :param path: path to search

    :return: list with all non-recursive files
    """
    return list(filter(lambda x: x.is_file(), path.iterdir()))


def get_all_files_with_suffix(path: pathlib.Path, suffix: str) -> list:
    """
    Get all non-recursive files of path

    :param path: path to search
    :param suffix: suffix to search

    :return: list with all non-recursive files
    """
    return list(filter(lambda x: x.suffix == suffix, path.iterdir()))


def get_datasplit_config_info_for_resultfolder(resultfolder: str) -> tuple:
    """
    Get all datasplit info for a result folder

    :param resultfolder: path to retrieve info

    :return: datasplit info with datasplit, n_outerfolds, n_innerfolds, val_set_size_percentage, test_set_size_percentage, maf_percentage
    """
    maf_perc = int(resultfolder.split('_')[2][3:])
    datasplit = resultfolder.split('_')[0]
    n_outerfolds = 1
    n_innerfolds = 5
    test_set_size_percentage = 20
    val_set_size_percentage = 20
    if datasplit == 'nested-cv':
        n_outerfolds = int(resultfolder.split('_')[1].split('-')[0])
        n_innerfolds = int(resultfolder.split('_')[1].split('-')[1])
    elif datasplit == 'cv-test':
        n_innerfolds = int(resultfolder.split('_')[1].split('-')[0])
        test_set_size_percentage = int(resultfolder.split('_')[1].split('-')[1])
    else:
        val_set_size_percentage = int(resultfolder.split('_')[1].split('-')[1][:-1])
        test_set_size_percentage = int(resultfolder.split('_')[1].split('-')[-1])
    return datasplit, n_outerfolds, n_innerfolds, val_set_size_percentage, test_set_size_percentage, maf_perc



import tensorflow as tf

from . import _tensorflow_model


class LocalCnn(_tensorflow_model.TensorflowModel):
    """
    Implementation of a class for a Locally-connected Convolutional Neural Network (LocalCNN).

    See :obj:`~easypheno.model._base_model.BaseModel` and :obj:`~easypheno.model._tensorflow_model.TensorflowModel` for more information on the attributes.
    """
    standard_encoding = 'onehot'
    possible_encodings = ['onehot']

    def define_model(self) -> tf.keras.Sequential:
        """
        Definition of a LocalCNN network.

        Architecture:

            - LocallyConnected1D, BatchNorm, Dropout, MaxPool1D, Flatten
            - N_LAYERS of (Dense + BatchNorm + Dropout)
            - Dense output layer

        Kernel size for LocallyConnectedLayer and max pooling layer may be fixed or optimized.
        Same applies for stride, number of units in the first dense layer and percentage decrease after each layer.
        """
        n_layers = self.suggest_hyperparam_to_optuna('n_layers')
        model = tf.keras.Sequential()
        act_function = tf.keras.layers.Activation(self.suggest_hyperparam_to_optuna('act_function'))
        l1_regularizer = None  # tf.keras.regularizers.L1(l1=self.suggest_hyperparam_to_optuna('l1_factor'))
        in_channels = self.width_onehot
        width = self.n_features
        model.add(tf.keras.Input(shape=(width, in_channels)))
        n_filters = 1
        kernel_size = int(2 ** self.suggest_hyperparam_to_optuna('kernel_size_exp'))
        stride = max(1, int(kernel_size * self.suggest_hyperparam_to_optuna('stride_perc_of_kernel_size')))
        model.add(tf.keras.layers.LocallyConnected1D(filters=n_filters, kernel_size=kernel_size,
                                                     strides=stride, activation=None,
                                                     kernel_regularizer=l1_regularizer))
        model.add(act_function)
        model.add(tf.keras.layers.BatchNormalization())
        p = self.suggest_hyperparam_to_optuna('dropout')
        model.add(tf.keras.layers.Dropout(rate=p, seed=42))
        kernel_size_max_pool = 2 ** 4  # self.suggest_hyperparam_to_optuna('maxpool_kernel_size_exp')
        model.add(tf.keras.layers.MaxPool1D(pool_size=kernel_size_max_pool))
        model.add(tf.keras.layers.Flatten())
        n_units = int(model.output_shape[1] * self.suggest_hyperparam_to_optuna('n_initial_units_factor'))
        perc_decrease = self.suggest_hyperparam_to_optuna('perc_decrease_per_layer')
        for layer in range(n_layers):
            model.add(tf.keras.layers.Dense(units=n_units, activation=None,
                                            kernel_regularizer=l1_regularizer))
            model.add(act_function)
            model.add(tf.keras.layers.BatchNormalization())
            model.add(tf.keras.layers.Dropout(rate=p))
            n_units = int(n_units * (1-perc_decrease))
        model.add(tf.keras.layers.Dense(units=self.n_outputs))
        return model

    def define_hyperparams_to_tune(self) -> dict:
        """
        See :obj:`~easypheno.model._base_model.BaseModel` for more information on the format.

        See TensorflowModel for more information on hyperparameters common for all tensorflow models.
        """

        if self.n_features < 20000:
            kernel_size_exp = {
                # Exponent with base 2 to get the kernel size for the convolutional layers
                'datatype': 'categorical',
                'list_of_values': [2.6, 3, 3.4, 3.6, 3.9]  # 6, 8, 10, 12, 14
            }
        else:
            kernel_size_exp = {
                # Exponent with base 2 to get the kernel size for the convolutional layers
                'datatype': 'int',
                'lower_bound': 3,
                'upper_bound': 7
            }

        return {
            'n_layers': {
                'datatype': 'int',
                'lower_bound': 1,
                'upper_bound': 3
            },
            'kernel_size_exp': kernel_size_exp,
            'maxpool_kernel_size_exp': {
                # Exponent with base 2 to get the kernel size for the maxpool layers
                'datatype': 'int',
                'lower_bound': 2,
                'upper_bound': 4
            },
            'stride_perc_of_kernel_size': {
                # Stride in relation to the kernel size
                'datatype': 'float',
                'lower_bound': 0.5,
                'upper_bound': 1,
                'step': 0.1
            },
            'n_initial_units_factor': {
                # Number of units in the linear layer after flattening in relation to the number of inputs
                'datatype': 'float',
                'lower_bound': 0.1,
                'upper_bound': 0.7,
                'step': 0.05
            },
            'perc_decrease_per_layer': {
                # Percentage decrease of the number of units per layer
                'datatype': 'float',
                'lower_bound': 0.2,
                'upper_bound': 0.5,
                'step': 0.05
            }
        }



import torch

from . import _torch_model


class Mlp(_torch_model.TorchModel):
    """
    Implementation of a class for a feedforward Multilayer Perceptron (MLP).

    See :obj:`~easypheno.model._base_model.BaseModel` and :obj:`~easypheno.model._torch_model.TorchModel` for more information on the attributes.
    """
    standard_encoding = '012'
    possible_encodings = ['012']

    def define_model(self) -> torch.nn.Sequential:
        """
        Definition of an MLP network.

        Architecture:

            - N_LAYERS of (Linear + BatchNorm + Dropout)
            - Linear output layer

        Number of units in the first linear layer and percentage decrease after each may be fixed or optimized.
        """
        n_layers = self.suggest_hyperparam_to_optuna('n_layers')
        model = []
        act_function = self.get_torch_object_for_string(string_to_get=self.suggest_hyperparam_to_optuna('act_function'))
        in_features = self.n_features
        out_features = int(in_features * self.suggest_hyperparam_to_optuna('n_initial_units_factor'))
        p = self.suggest_hyperparam_to_optuna('dropout')
        perc_decrease = self.suggest_hyperparam_to_optuna('perc_decrease_per_layer')
        for layer in range(n_layers):
            model.append(torch.nn.Linear(in_features=in_features, out_features=out_features))
            model.append(act_function)
            model.append(torch.nn.BatchNorm1d(num_features=out_features))
            model.append(torch.nn.Dropout(p=p))
            in_features = out_features
            out_features = int(in_features * (1-perc_decrease))
        model.append(torch.nn.Linear(in_features=in_features, out_features=self.n_outputs))
        return torch.nn.Sequential(*model)

    def define_hyperparams_to_tune(self) -> dict:
        """
        See :obj:`~easypheno.model._base_model.BaseModel` for more information on the format.

        See :obj:`~easypheno.model._torch_model.TorchModel` for more information on hyperparameters common for all torch models.
        """
        n_layers = {
            'datatype': 'int',
            'lower_bound': 1,
            'upper_bound': 5
        }
        n_initial_units_factor = {
            # Number of units in the first linear layer in relation to the number of inputs
            'datatype': 'float',
            'lower_bound': 0.1,
            'upper_bound': 0.7,
            'step': 0.05
        }
        perc_decrease_per_layer = {
            # Percentage decrease of the number of units per layer
            'datatype': 'float',
            'lower_bound': 0.1,
            'upper_bound': 0.5,
            'step': 0.05
        }
        if self.n_features > 20000:
            n_layers = {
                'datatype': 'int',
                'lower_bound': 1,
                'upper_bound': 5
            }
            n_initial_units_factor = {
                # Number of units in the first linear layer in relation to the number of inputs
                'datatype': 'float',
                'lower_bound': 0.1,
                'upper_bound': 0.3,
                'step': 0.01
            }
            perc_decrease_per_layer = {
                # Percentage decrease of the number of units per layer
                'datatype': 'float',
                'lower_bound': 0.1,
                'upper_bound': 0.5,
                'step': 0.05
            }
        if self.n_features > 50000:
            n_layers = {
                'datatype': 'int',
                'lower_bound': 1,
                'upper_bound': 3
            }
            n_initial_units_factor = {
                # Number of units in the first linear layer in relation to the number of inputs
                'datatype': 'float',
                'lower_bound': 0.01,
                'upper_bound': 0.15,
                'step': 0.01
            }
            perc_decrease_per_layer = {
                # Percentage decrease of the number of units per layer
                'datatype': 'float',
                'lower_bound': 0.2,
                'upper_bound': 0.5,
                'step': 0.05
            }

        return {
            'n_layers': n_layers,
            'n_initial_units_factor': n_initial_units_factor,
            'perc_decrease_per_layer': perc_decrease_per_layer
        }



import argparse

from . import results_analysis

if __name__ == "__main__":
    """
    Run to gather some overview files on the optimization results for the specified results directory
    """
    parser = argparse.ArgumentParser()
    parser.add_argument("-rd", "--results_dir", type=str,
                        help="Provide the full path of the directory where your results are stored "
                             "(name of the genotype matrix level)")
    parser.add_argument("-evm", "--eval_metric", type=str,
                        help="Eval metric to use for results summary. Options, default given first: "
                             "Regression: explained_variance, r2_score, rmse, mse. "
                             "Classification: mcc, f1_score, accuracy, precision, recall.")
    args = vars(parser.parse_args())
    results_directory_genotype_level = args['results_dir']
    eval_metric = args['eval_metric']

    results_analysis.summarize_results_per_phenotype_and_datasplit(
        results_directory_genotype_level=results_directory_genotype_level, eval_metric=eval_metric
    )



import abc
import numpy as np
import optuna
import torch.nn
import torch.utils.data
import copy

from . import _base_model


class TorchModel(_base_model.BaseModel, abc.ABC):
    """
    Parent class based on :obj:`~easypheno.model._base_model.BaseModel` for all PyTorch models to share functionalities.
    See :obj:`~easypheno.model._base_model.BaseModel` for more information.

    *Attributes*

        *Inherited attributes*

        See :obj:`~easypheno.model._base_model.BaseModel`.

        *Additional attributes*

        - n_features (*int*): Number of input features to the model
        - width_onehot (*int*): Number of input channels in case of onehot encoding
        - batch_size (*int*): Batch size for batch-based training
        - n_epochs (*int*): Number of epochs for optimization
        - optimizer (*torch.optim.optimizer.Optimizer*): optimizer for model fitting
        - loss_fn: loss function for model fitting
        - early_stopping_patience (*int*): epochs without improvement before early stopping
        - early_stopping_point (*int*): epoch at which early stopping occured
        - device (*torch.device*): device to use, e.g. GPU

    :param task: ML task (regression or classification) depending on target variable
    :param optuna_trial: optuna.trial.Trial : trial of optuna for optimization
    :param encoding: the encoding to use (standard encoding or user-defined)
    :param n_outputs: Number of outputs of the model
    :param n_features: Number of input features to the model
    :param width_onehot: Number of input channels in case of onehot encoding
    :param batch_size: Batch size for batch-based training
    :param n_epochs: Number of epochs for optimization
    :param early_stopping_point: Stop training at defined epoch
    """

    def __init__(self, task: str, optuna_trial: optuna.trial.Trial, encoding: str = None, n_outputs: int = 1,
                 n_features: int = None, width_onehot: int = None, batch_size: int = None, n_epochs: int = None,
                 early_stopping_point: int = None):
        self.all_hyperparams = self.common_hyperparams()  # add hyperparameters commonly optimized for all torch models
        self.n_features = n_features
        self.width_onehot = width_onehot  # relevant for models using onehot encoding e.g. CNNs
        super().__init__(task=task, optuna_trial=optuna_trial, encoding=encoding, n_outputs=n_outputs)
        self.batch_size = \
            batch_size if batch_size is not None else self.suggest_hyperparam_to_optuna('batch_size')
        self.n_epochs = n_epochs if n_epochs is not None else self.suggest_hyperparam_to_optuna('n_epochs')
        self.optimizer = torch.optim.Adam(params=self.model.parameters(),
                                          lr=self.suggest_hyperparam_to_optuna('learning_rate'))
        self.loss_fn = torch.nn.CrossEntropyLoss() if task == 'classification' else torch.nn.MSELoss()
        # self.l1_factor = self.suggest_hyperparam_to_optuna('l1_factor')
        # early stopping if there is no improvement on validation loss for a certain number of epochs
        self.early_stopping_patience = self.suggest_hyperparam_to_optuna('early_stopping_patience')
        self.early_stopping_point = early_stopping_point
        self.device = torch.device('cuda:0' if torch.cuda.is_available() else 'cpu')

    def train_val_loop(self, X_train: np.array, y_train: np.array, X_val: np.array, y_val: np.array) -> np.array:
        """
        Implementation of a train and validation loop for  PyTorch models.
        See :obj:`~easypheno.model._base_model.BaseModel` for more information
        """
        train_loader = self.get_dataloader(X=X_train, y=y_train)
        val_loader = self.get_dataloader(X=X_val, y=y_val)
        best_model = copy.deepcopy(self.model)
        self.model.to(device=self.device)
        best_loss = None
        epochs_wo_improvement = 0
        for epoch in range(self.n_epochs):
            self.train_one_epoch(train_loader=train_loader)
            val_loss = self.validate_one_epoch(val_loader=val_loader)
            if best_loss is None or val_loss < best_loss:
                best_loss = val_loss
                epochs_wo_improvement = 0
                best_model = copy.deepcopy(self.model)
            else:
                epochs_wo_improvement += 1
            print('Epoch ' + str(epoch + 1) + ' of ' + str(self.n_epochs))
            print('Current val_loss=' + str(val_loss) + ', best val_loss=' + str(best_loss))
            if epoch >= 20 and epochs_wo_improvement >= self.early_stopping_patience:
                print("Early Stopping at " + str(epoch + 1) + ' of ' + str(self.n_epochs))
                self.early_stopping_point = epoch - self.early_stopping_patience
                self.model = best_model
                return self.predict(X_in=X_val)
        return self.predict(X_in=X_val)

    def train_one_epoch(self, train_loader: torch.utils.data.DataLoader):
        """
        Train one epoch

        :param train_loader: DataLoader with training data
        """
        self.model.train()
        for inputs, targets in train_loader:
            inputs, targets = inputs.to(device=self.device), targets.to(device=self.device)
            self.optimizer.zero_grad()
            outputs = self.model(inputs)
            loss = self.get_loss(outputs=outputs, targets=targets)
            # l1_loss = 0
            # for param in self.model.parameters():
            #     l1_loss += torch.sum(torch.abs(param))
            # loss += self.l1_factor * l1_loss
            loss.backward()
            self.optimizer.step()

    def validate_one_epoch(self, val_loader: torch.utils.data.DataLoader) -> float:
        """
        Validate one epoch

        :param val_loader: DataLoader with validation data

        :return: loss based on loss-criterion
        """
        self.model.eval()
        total_loss = 0
        with torch.no_grad():
            for inputs, targets in val_loader:
                inputs, targets = inputs.to(device=self.device), targets.to(device=self.device)
                outputs = self.model(inputs)
                total_loss += self.get_loss(outputs=outputs, targets=targets).item()
        return total_loss / len(val_loader.dataset)

    def retrain(self, X_retrain: np.array, y_retrain: np.array):
        """
        Implementation of the retraining for PyTorch models.
        See :obj:`~easypheno.model._base_model.BaseModel` for more information
        """
        retrain_loader = self.get_dataloader(X=X_retrain, y=y_retrain)
        n_epochs_to_retrain = self.n_epochs if self.early_stopping_point is None else self.early_stopping_point
        self.model.to(device=self.device)
        for epoch in range(n_epochs_to_retrain):
            print('Retrain: Epoch ' + str(epoch + 1) + ' of ' + str(n_epochs_to_retrain))
            self.train_one_epoch(retrain_loader)

    def predict(self, X_in: np.array) -> np.array:
        """
        Implementation of a prediction based on input features for PyTorch models.
        See :obj:`~easypheno.model._base_model.BaseModel` for more information
        """
        dataloader = self.get_dataloader(X=X_in, shuffle=False)
        self.model.eval()
        predictions = None
        with torch.no_grad():
            for inputs in dataloader:
                inputs = inputs.to(device=self.device)
                outputs = self.model(inputs)
                predictions = torch.clone(outputs) if predictions is None else torch.cat((predictions, outputs))
        if self.task == 'classification':
            _, predictions = torch.max(predictions, 1)
        return predictions.cpu().detach().numpy()

    def get_loss(self, outputs: torch.Tensor, targets: torch.Tensor) -> torch.Tensor:
        """
        Calculate the loss based on the outputs and targets

        :param outputs: outputs of the model
        :param targets: targets of the dataset

        :return: loss
        """
        if type(self.loss_fn) in [torch.nn.CrossEntropyLoss, torch.nn.NLLLoss]:
            targets = targets.long()
        return self.loss_fn(outputs, targets)

    def get_dataloader(self, X: np.array, y: np.array = None, shuffle: bool = True) -> torch.utils.data.DataLoader:
        """
        Get a Pytorch DataLoader using the specified data and batch size

        :param X: feature matrix to use
        :param y: optional target vector to use
        :param shuffle: shuffle parameter for DataLoader

        :return: Pytorch DataLoader
        """
        # drop last sample if last batch would only contain one sample
        if (len(X) % self.batch_size) == 1:
            X = X[:-1]
            y = y[:-1] if y is not None else None
        X = torch.from_numpy(X).float()
        if self.encoding == 'onehot':
            # Adapt to PyTorch ordering (BATCH_SIZE, CHANNELS, SIGNAL)
            X = torch.swapaxes(X, 1, 2)
        y = torch.reshape(torch.from_numpy(y).float(), (-1, 1)) if y is not None else None
        y = torch.flatten(y) if (self.task == 'classification' and y is not None) else y
        dataset = torch.utils.data.TensorDataset(X, y) if y is not None \
            else X
        return torch.utils.data.DataLoader(dataset=dataset, batch_size=self.batch_size, shuffle=shuffle)

    @staticmethod
    def common_hyperparams():
        """
        Add hyperparameters that are common for PyTorch models.
        Do not need to be included in optimization for every child model.
        Also See :obj:`~easypheno.model._base_model.BaseModel` for more information
        """
        return {
            'dropout': {
                'datatype': 'float',
                'lower_bound': 0,
                'upper_bound': 0.5,
                'step': 0.1
            },
            'act_function': {
                'datatype': 'categorical',
                'list_of_values': ['relu', 'tanh']
            },
            'batch_size': {
                'datatype': 'categorical',
                'list_of_values': [4, 8, 16, 32, 64]
            },
            'n_epochs': {
                'datatype': 'categorical',
                'list_of_values': [100, 500, 1000, 5000, 10000]
            },
            'learning_rate': {
                'datatype': 'categorical',
                'list_of_values': [1e-6, 1e-5, 1e-4, 1e-3, 1e-2, 1e-1]
            },
            'early_stopping_patience': {
                'datatype': 'int',
                'lower_bound': 0,
                'upper_bound': 20,
                'step': 5
            },
            'l1_factor': {
                'datatype': 'float',
                'lower_bound': 0,
                'upper_bound': 10**3
            }
        }

    @staticmethod
    def get_torch_object_for_string(string_to_get: str):
        """
        Get the torch object for a specific string, e.g. when suggesting to optuna as hyperparameter

        :param string_to_get: string to retrieve the torch object

        :return: torch object
        """
        string_to_object_dict = {
            'relu': torch.nn.ReLU(),
            'tanh': torch.nn.Tanh()
        }
        return string_to_object_dict[string_to_get]



import h5py
import numpy as np
import sklearn.preprocessing
import pathlib

from ..utils import helper_functions
from . import raw_data_functions, encoding_functions


class Dataset:
    """
    Class containing dataset ready for optimization (e.g. geno/phenotype matched).

    **Attributes**

        - encoding (*str*): the encoding to use (standard encoding or user-defined)
        - X_full (*numpy.array*): all (matched, maf- and duplicated-filtered) SNPs
        - y_full (*numpy.array*): all target values
        - sample_ids_full (*numpy.array*):all sample ids
        - snp_ids (*numpy.array*): SNP ids
        - datasplit (*str*): datasplit to use
        - datasplit_indices (*dict*): dictionary containing all indices for the specified datasplit

    :param data_dir: data directory where the phenotype and genotype matrix are stored
    :param genotype_matrix_name: name of the genotype matrix including datatype ending
    :param phenotype_matrix_name: name of the phenotype matrix including datatype ending
    :param phenotype: name of the phenotype to predict
    :param datasplit: datasplit to use. Options are: nested-cv, cv-test, train-val-test
    :param n_outerfolds: number of outerfolds relevant for nested-cv
    :param n_innerfolds: number of folds relevant for nested-cv and cv-test
    :param test_set_size_percentage: size of the test set relevant for cv-test and train-val-test
    :param val_set_size_percentage: size of the validation set relevant for train-val-test
    :param maf_percentage: threshold for MAF filter as percentage value
    :param encoding: the encoding to use (standard encoding or user-defined)
    :param do_snp_filters: specify if SNP filters (e.g. duplicates, maf etc.) should be applied
    """

    def __init__(self, data_dir: pathlib.Path, genotype_matrix_name: str, phenotype_matrix_name: str, phenotype: str,
                 datasplit: str, n_outerfolds: int, n_innerfolds: int, test_set_size_percentage: int,
                 val_set_size_percentage: int, encoding: str, maf_percentage: int, do_snp_filters: bool = True):
        self.encoding = encoding
        self.datasplit = datasplit
        self.index_file_name = self.get_index_file_name(
                genotype_matrix_name=genotype_matrix_name, phenotype_matrix_name=phenotype_matrix_name,
                phenotype=phenotype
        )
        self.X_full, self.y_full, self.sample_ids_full, self.snp_ids = self.load_match_raw_data(
            data_dir=data_dir, genotype_matrix_name=genotype_matrix_name, do_snp_filters=do_snp_filters)
        if do_snp_filters:
            self.maf_filter_raw_data(data_dir=data_dir, maf_percentage=maf_percentage)
            self.filter_duplicate_snps()
        self.check_and_save_filtered_snp_ids(data_dir=data_dir, maf_percentage=maf_percentage)
        self.datasplit_indices = self.load_datasplit_indices(
            data_dir=data_dir, n_outerfolds=n_outerfolds, n_innerfolds=n_innerfolds,
            test_set_size_percentage=test_set_size_percentage, val_set_size_percentage=val_set_size_percentage
        )
        self.check_datasplit(
            n_outerfolds=1 if datasplit != 'nested-cv' else n_outerfolds,
            n_innerfolds=1 if datasplit == 'train-val-test' else n_innerfolds
        )

    def load_match_raw_data(self, data_dir: pathlib.Path, genotype_matrix_name: str, do_snp_filters: bool = True) \
            -> (np.ndarray, np.ndarray, np.ndarray):
        """
        Load the full genotype and phenotype matrices specified and match them

        :param data_dir: data directory where the phenotype and genotype matrix are stored
        :param genotype_matrix_name: name of the genotype matrix including datatype ending
        :param do_snp_filters: specify if SNP filters (e.g. duplicates, maf etc.) should be applied

        :return: matched genotype, phenotype and sample ids
        """
        print('Load and match raw data')
        with h5py.File(data_dir.joinpath(self.index_file_name), "r") as f:
            # Load information from index file
            y = f['matched_data/y'][:]  # TODO change if multiple phenotypes
            if helper_functions.test_likely_categorical(y):
                if y.dtype.type is np.float64:
                    y = y.astype(int)
                y = sklearn.preprocessing.LabelEncoder().fit_transform(y)
            sample_ids = f['matched_data/matched_sample_ids'][:].astype(str)
            non_informative_filter = f['matched_data/non_informative_filter'][:]
            X_index = f['matched_data/X_index'][:]

        with h5py.File(data_dir.joinpath(genotype_matrix_name).with_suffix('.h5'), "r") as f:
            # Load genotype data
            snp_ids = f['snp_ids'][non_informative_filter].astype(str) if do_snp_filters \
                else f['snp_ids'][:].astype(str)
            if f'X_{self.encoding}' in f:
                X = f[f'X_{self.encoding}'][:, non_informative_filter] if do_snp_filters else f[f'X_{self.encoding}'][:]
            elif f'X_{encoding_functions.get_base_encoding(encoding=self.encoding)}' in f:
                X_base = \
                    f[f'X_{encoding_functions.get_base_encoding(encoding=self.encoding)}'][:, non_informative_filter] \
                        if do_snp_filters else f[f'X_{encoding_functions.get_base_encoding(encoding=self.encoding)}'][:]
                X = encoding_functions.encode_genotype(X=X_base, required_encoding=self.encoding)
            else:
                raise Exception('Genotype in ' + self.encoding + ' encoding missing. Can not create required encoding. '
                                                                 'See documentation for help')
        X = raw_data_functions.get_matched_data(data=X, index=X_index)
        # sanity checks
        raw_data_functions.check_genotype_shape(X=X, sample_ids=sample_ids, snp_ids=snp_ids)
        if raw_data_functions.check_duplicate_samples(sample_ids=sample_ids):
            raise Exception('The genotype matrix contains duplicate samples. Please check again.')
        return X, np.reshape(y, (-1, 1)), np.reshape(sample_ids, (-1, 1)), snp_ids

    def maf_filter_raw_data(self, data_dir: pathlib.Path, maf_percentage: int):
        """
        Apply maf filter to full raw data, if maf=0 only non-informative SNPs will be removed

        :param data_dir: data directory where the phenotype and genotype matrix are stored
        :param maf_percentage: threshold for MAF filter as percentage value
        """
        print('Apply MAF filter')
        with h5py.File(data_dir.joinpath(self.index_file_name), "r") as f:
            if f'maf_filter/maf_{maf_percentage}' in f:
                filter_indices = f[f'maf_filter/maf_{maf_percentage}'][:]
        self.X_full = np.delete(self.X_full, filter_indices, axis=1)
        self.snp_ids = np.delete(self.snp_ids, filter_indices)

    def filter_duplicate_snps(self):
        """
        Remove duplicate SNPs,
        i.e. SNPs that are completely the same for all samples and therefore do not add information.
        """
        print('Filter duplicate SNPs')
        uniques, index = np.unique(self.X_full, return_index=True, axis=1)
        self.X_full = self.X_full[:, np.sort(index)]
        self.snp_ids = self.snp_ids[np.sort(index)]

    def check_and_save_filtered_snp_ids(self, data_dir: pathlib.Path, maf_percentage: int):
        """
        Check if snp_ids for specific maf percentage and encoding are saved in index_file.
        If not, save them in 'matched_data/final_snp_ids/{encoding}/maf_{maf_percentage}_snp_ids'

        :param data_dir: data directory where the phenotype and genotype matrix are stored
        :param maf_percentage: threshold for MAF filter as percentage value
        """
        print('Check if final snp_ids already exist in index_file for used encoding and maf percentage. '
              'Save them if necessary.')
        with h5py.File(data_dir.joinpath(self.index_file_name), "a") as f:
            if 'final_snp_ids' not in f['matched_data']:
                final = f.create_group('matched_data/final_snp_ids')
                enc = final.create_group(f'{self.encoding}')
                enc.create_dataset(f'maf_{maf_percentage}_snp_ids',
                                   data=self.snp_ids.astype(bytes), chunks=True, compression="gzip")
            elif f'{self.encoding}' not in f['matched_data/final_snp_ids']:
                enc = f.create_group(f'matched_data/final_snp_ids/{self.encoding}')
                enc.create_dataset(f'maf_{maf_percentage}_snp_ids',
                                   data=self.snp_ids.astype(bytes), chunks=True, compression="gzip")
            elif f'maf_{maf_percentage}_snp_ids' not in f[f'matched_data/final_snp_ids/{self.encoding}']:
                f.create_dataset(f'matched_data/final_snp_ids/{self.encoding}/maf_{maf_percentage}_snp_ids',
                                 data=self.snp_ids.astype(bytes), chunks=True, compression="gzip")

    def load_datasplit_indices(self, data_dir: pathlib.Path, n_outerfolds: int, n_innerfolds: int,
                               test_set_size_percentage: int, val_set_size_percentage: int) -> dict:
        """
        Load the datasplit indices saved during file unification.

        Structure:

        .. code-block:: python

            {
                'outerfold_0': {
                    'innerfold_0': {'train': indices_train, 'val': indices_val},
                    'innerfold_1': {'train': indices_train, 'val': indices_val},
                    ...
                    'innerfold_n': {'train': indices_train, 'val': indices_val},
                    'test': test_indices
                    },
                ...
                'outerfold_m': {
                    'innerfold_0': {'train': indices_train, 'val': indices_val},
                    'innerfold_1': {'train': indices_train, 'val': indices_val},
                    ...
                    'innerfold_n': {'train': indices_train, 'val': indices_val},
                    'test': test_indices
                    }
            }

        Caution: The actual structure depends on the datasplit specified by the user,
        e.g. for a train-val-test split only 'outerfold_0' and its subelements 'innerfold_0' and 'test' exist.

        :param data_dir: data directory where the phenotype and genotype matrix are stored
        :param n_outerfolds: number of outerfolds relevant for nested-cv
        :param n_innerfolds: number of folds relevant for nested-cv and cv-test
        :param test_set_size_percentage: size of the test set relevant for cv-test and train-val-test
        :param val_set_size_percentage: size of the validation set relevant for train-val-test

        :return: dictionary with the above-described structure containing all indices for the specified data split
        """
        print('Load datasplit file')
        # construct variables for further process
        if self.datasplit == 'train-val-test':
            n_outerfolds = 1
            n_innerfolds = 1
            datasplit_params = [val_set_size_percentage, test_set_size_percentage]
        elif self.datasplit == 'cv-test':
            n_outerfolds = 1
            n_innerfolds = n_innerfolds
            datasplit_params = [n_innerfolds, test_set_size_percentage]
        elif self.datasplit == 'nested-cv':
            n_outerfolds = n_outerfolds
            n_innerfolds = n_innerfolds
            datasplit_params = [n_outerfolds, n_innerfolds]
        split_param_string = helper_functions.get_subpath_for_datasplit(datasplit=self.datasplit,
                                                                        datasplit_params=datasplit_params)

        datasplit_indices = {}
        with h5py.File(data_dir.joinpath(self.index_file_name), "r") as f:
            # load datasplit indices from index file to ensure comparability between different models
            for m in range(n_outerfolds):
                outerfold_path = \
                    f'datasplits/{self.datasplit}/{split_param_string}/outerfold_{m}/'
                datasplit_indices['outerfold_' + str(m)] = {'test': f[f'{outerfold_path}test/'][:]}
                for n in range(n_innerfolds):
                    datasplit_indices['outerfold_' + str(m)]['innerfold_' + str(n)] = \
                        {
                            'train': f[f'{outerfold_path}innerfold_{n}/train'][:],
                            'val': f[f'{outerfold_path}innerfold_{n}/val'][:]
                        }
        return datasplit_indices

    def check_datasplit(self, n_outerfolds: int, n_innerfolds: int):
        """
        Check if the datasplit is valid. Raise Exceptions if train, val or test sets contain same samples.

        :param n_outerfolds: number of outerfolds in datasplit_indices dictionary
        :param n_innerfolds: number of folds in datasplit_indices dictionary
        """
        all_sample_ids_test = []
        for j in range(n_outerfolds):
            sample_ids_test = set(self.sample_ids_full[self.datasplit_indices[f'outerfold_{j}']['test']].flatten())
            all_sample_ids_test.extend(sample_ids_test)
            for i in range(n_innerfolds):
                sample_ids_train = set(
                    self.sample_ids_full[self.datasplit_indices[f'outerfold_{j}'][f'innerfold_{i}']['train']].flatten()
                )
                sample_ids_val = set(
                    self.sample_ids_full[self.datasplit_indices[f'outerfold_{j}'][f'innerfold_{i}']['val']].flatten())
                if len(sample_ids_train.intersection(sample_ids_val)) != 0:
                    raise Exception(
                        'Something with the datasplit went wrong - the intersection of train and val samples is not '
                        'empty. Please check again.'
                    )
                if len(sample_ids_train.intersection(sample_ids_test)) != 0:
                    raise Exception(
                        'Something with the datasplit went wrong - the intersection of train and test samples is not '
                        'empty. Please check again.'
                    )
                if len(sample_ids_val.intersection(sample_ids_test)) != 0:
                    raise Exception(
                        'Something with the datasplit went wrong - the intersection of val and test samples is not '
                        'empty. Please check again.'
                    )
        if self.datasplit == 'nested-cv':
            if len(set(all_sample_ids_test).intersection(set(self.sample_ids_full.flatten()))) \
                    != len(set(self.sample_ids_full.flatten())):
                raise Exception('Something with the datasplit went wrong - '
                                'not all sample ids are in one of the outerfold test sets')
        print('Checked datasplit for all folds.')

    @staticmethod
    def get_index_file_name(genotype_matrix_name: str, phenotype_matrix_name: str, phenotype: str) -> str:
        """
        Get the name of the file containing the indices for maf filters and data splits

        :param genotype_matrix_name: name of the genotype matrix including datatype ending
        :param phenotype_matrix_name: name of the phenotype matrix including datatype ending
        :param phenotype: name of the phenotype to predict

        :return: name of index file
        """
        return genotype_matrix_name.split('.')[0] + '-' + phenotype_matrix_name.split('.')[0] + '-' + phenotype + '.h5'



import pathlib
import pandas as pd
import numpy as np

from . import helper_functions
from ..model import _param_free_base_model, _torch_model, _tensorflow_model


def check_all_specified_arguments(arguments: dict):
    """
    Check all specified arguments for plausibility

    :param arguments: all arguments provided by the user
    """
    # Check existence of save_dir
    if not arguments["save_dir"].exists():
        raise Exception("Specified save_dir " + str(arguments["save_dir"]) + " does not exist. Please double-check.")
    # Check existence of genotype and phenotype file
    if not arguments["data_dir"].joinpath(arguments["genotype_matrix"]).is_file():
        raise Exception('Specified genotype file ' + arguments["genotype_matrix"] + ' does not exist in '
                        + str(arguments["data_dir"]) + '. Please check spelling.')
    phenotype_file = arguments["data_dir"].joinpath(arguments["phenotype_matrix"])
    if not phenotype_file.is_file():
        raise Exception('Specified phenotype file ' + arguments["phenotype_matrix"] + ' does not exist in '
                        + str(arguments["data_dir"]) + '. Please check spelling.')
    # Check existence of specified phenotype in phenotype file
    phenotype = pd.read_csv(phenotype_file)
    if arguments["phenotype"] not in phenotype.columns:
        raise Exception('Specified phenotype ' + arguments["phenotype"] + ' does not exist in phenotype file '
                        + str(phenotype_file) + '. Check spelling.')

    # Check meaningfulness of specified values
    if not (0 <= arguments["maf_percentage"] <= 20):
        raise Exception('Specified maf value of ' + str(arguments["maf_percentage"]) 
                        + ' is invalid, has to be between 0 and 20.')
    if not (5 <= arguments["test_set_size_percentage"] <= 30):
        raise Exception('Specified test set size in percentage ' + str(arguments["test_set_size_percentage"]) +
                        ' is invalid, has to be between 5 and 30.')
    if not (5 <= arguments["val_set_size_percentage"] <= 30):
        raise Exception('Specified validation set size in percentage ' 
                        + str(arguments["val_set_size_percentage"]) + ' is invalid, has to be between 5 and 30.')
    if not (3 <= arguments["n_outerfolds"] <= 10):
        raise Exception('Specified number of outerfolds ' + str(arguments["n_outerfolds"]) +
                        ' is invalid, has to be between 3 and 10.')
    if not (3 <= arguments["n_innerfolds"] <= 10):
        raise Exception('Specified number of innerfolds/folds ' + str(arguments["n_innerfolds"]) +
                        ' is invalid, has to be between 3 and 10.')
    if "n_trials" in arguments and any([not issubclass(helper_functions.get_mapping_name_to_class()[model],
                                                       _param_free_base_model.ParamFreeBaseModel)
                                        for model in arguments["models"]]) and arguments["n_trials"] < 10:
        raise Exception('Specified number of trials with ' + str(arguments["n_trials"]) + ' is invalid, at least 10.')

    # Check spelling of datasplit and model
    if arguments["datasplit"] not in ['nested-cv', 'cv-test', 'train-val-test']:
        raise Exception('Specified datasplit ' + arguments["datasplit"] + ' is invalid, '
                        'has to be: nested-cv | cv-test | train-val-test')
    if (arguments["models"] != 'all') and \
            (any(model not in helper_functions.get_list_of_implemented_models() for model in arguments["models"])):
        raise Exception('At least one specified model in "' + str(arguments["models"]) +
                        '" not found in implemented models nor "all" specified.' +
                        ' Check spelling or if implementation exists. Implemented models: ' +
                        str(helper_functions.get_list_of_implemented_models()))

    # Check encoding
    if arguments["encoding"] is not None:
        if arguments["encoding"] not in ['raw', '012', 'onehot']:
            raise Exception('Specified encoding ' + arguments["encoding"] + ' is not valid. See help.')
        else:
            if arguments["models"] == 'all' or len(arguments["models"]) > 1:
                raise Exception('If "all" models are specified, standard encodings are used. Do not specify encoding')
            else:
                if arguments["encoding"] not in \
                        helper_functions.get_mapping_name_to_class()[arguments["models"][0]].possible_encodings:
                    raise Exception(arguments["encoding"] + ' is not valid for ' + arguments["models"][0] +
                                    '. Check possible_encodings in model file.')

    # Only relevant for neural networks
    if any([issubclass(helper_functions.get_mapping_name_to_class()[model], _torch_model.TorchModel) or \
            issubclass(helper_functions.get_mapping_name_to_class()[model] ,  _tensorflow_model.TensorflowModel)
            for model in arguments["models"]]):
        if arguments["batch_size"] is not None:
            if not (2**3 <= arguments["batch_size"] <= 2**8):
                raise Exception('Specified batch size ' + str(arguments["batch_size"]) +
                                ' is invalid, has to be between 8 and 256.')
        if arguments["n_epochs"] is not None:
            if not (50 <= arguments["n_epochs"] <= 1000000):
                raise Exception('Specified number of epochs ' + str(arguments["n_epochs"]) +
                                ' is invalid, has to be between 50 and 1.000.000.')


def check_exist_directories(list_of_dirs: list, create_if_not_exist: bool = False) -> bool:
    """
    Check if each directory within a list exists

    :param list_of_dirs: list with directories as pathlib.Path
    :param create_if_not_exist: bool if non-existing directories should be created

    :return: True if all exist, False otherwise
    """
    check = True
    for dir_to_check in list_of_dirs:
        if not dir_to_check.exists():
            print("Directory " + str(dir_to_check) + " not existing.")
            if create_if_not_exist:
                print("Will create it.")
                dir_to_check.mkdir(parents=True)
            else:
                print("Please correct it.")
                check = False
    return check


def check_exist_files(list_of_files: list) -> bool:
    """
    Check if each file within a list exists

    :param list_of_files: list with files as pathlib.Path

    :return: True if all exist, False otherwise
    """
    check = True
    for file_to_check in list_of_files:
        if not file_to_check.is_file():
            print("File " + str(file_to_check) + " not existing.")
            print("Please correct it.")
            check = False
    return check


def compare_snp_id_vectors(snp_id_vector_small_equal: np.array, snp_id_vector_big_equal: np.array) -> bool:
    """
    Compare two SNP id vectors if they contain the same ids

    :param snp_id_vector_small_equal: vector 1 with SNP ids, can be a (smaller) subset of snp_id_vector_big_equal
    :param snp_id_vector_big_equal: vector 2 with SNP ids, can contain more SNP ids than snp_id_vector_small_equal

    :return: True if snp_id_vector_small_equal is a subset of the other vector
    """
    return set(snp_id_vector_small_equal).issubset(set(snp_id_vector_big_equal))



import datetime
import optuna
import pandas as pd
import sklearn
import sklearn.inspection
import numpy as np
import shutil
import time
import csv
import gc
import pathlib

import torch.cuda
import tensorflow as tf

from ..preprocess import base_dataset
from ..utils import helper_functions
from ..evaluation import eval_metrics
from ..model import _base_model, _model_functions, _tensorflow_model, _torch_model


class OptunaOptim:
    """
    Class that contains all info for the whole optimization using optuna for one model and dataset.

    **Attributes**

        - task (*str*): ML task (regression or classification) depending on target variable
        - current_model_name (*str*): name of the current model according to naming of .py file in package model
        - dataset (:obj:`~easypheno.preprocess.base_dataset.Dataset`): dataset to use for optimization run
        - datasplit_subpath (*str*): subpath with datasplit info relevant for saving / naming
        - base_path (*str*): base_path for save_path
        - save_path (*str*): path for model and results storing
        - study (*optuna.study.Study*): optuna study for optimization run
        - current_best_val_result (*float*): the best validation result so far
        - early_stopping_point (*int*): point at which early stopping occured (relevant for some models)
        - user_input_params (*dict*): all params handed over to the constructor that are needed in the whole class

    :param save_dir: directory for saving the results.
    :param genotype_matrix_name: name of the genotype matrix including datatype ending
    :param phenotype_matrix_name: name of the phenotype matrix including datatype ending
    :param phenotype: name of the phenotype to predict
    :param n_outerfolds: number of outerfolds relevant for nested-cv
    :param n_innerfolds: number of folds relevant for nested-cv and cv-test
    :param test_set_size_percentage: size of the test set relevant for cv-test and train-val-test
    :param val_set_size_percentage: size of the validation set relevant for train-val-test
    :param maf_percentage: threshold for MAF filter as percentage value
    :param n_trials: number of trials for optuna
    :param save_final_model: specify if the final model should be saved
    :param batch_size: batch size for neural network models
    :param n_epochs: number of epochs for neural network models
    :param task: ML task (regression or classification) depending on target variable
    :param current_model_name: name of the current model according to naming of .py file in package model
    :param dataset: dataset to use for optimization run
    :param models_start_time: optimized models and starting time of the optimization run for saving purposes
    :param intermediate_results_interval: number of trials after which intermediate results will be saved
    :param outerfold_number_to_run: outerfold to run in case you do not want to run all
    """

    def __init__(self, save_dir: pathlib.Path, genotype_matrix_name: str, phenotype_matrix_name: str, phenotype: str,
                 n_outerfolds: int, n_innerfolds: int, val_set_size_percentage: int, test_set_size_percentage: int,
                 maf_percentage: int, n_trials: int, save_final_model: bool, batch_size: int, n_epochs: int,
                 task: str, current_model_name: str, dataset: base_dataset.Dataset, models_start_time: str,
                 intermediate_results_interval: int = 50, outerfold_number_to_run: int = None):
        self.current_model_name = current_model_name
        self.task = task
        self.dataset = dataset
        if self.dataset.datasplit == 'train-val-test':
            datasplit_params = [val_set_size_percentage, test_set_size_percentage]
        elif self.dataset.datasplit == 'cv-test':
            datasplit_params = [n_innerfolds, test_set_size_percentage]
        elif self.dataset.datasplit == 'nested-cv':
            datasplit_params = [n_outerfolds, n_innerfolds]
        self.datasplit_subpath = helper_functions.get_subpath_for_datasplit(
            datasplit=self.dataset.datasplit, datasplit_params=datasplit_params
        )
        self.base_path = save_dir.joinpath('results', genotype_matrix_name.split('.')[0], \
                         phenotype_matrix_name.split('.')[0], phenotype, self.dataset.datasplit + '_' + \
                         self.datasplit_subpath + '_MAF' + str(maf_percentage) + '_' + models_start_time, \
                         current_model_name)
        self.save_path = self.base_path
        self.study = None
        self.current_best_val_result = None
        self.early_stopping_point = None
        self.intermediate_results_interval = intermediate_results_interval
        self.user_input_params = locals()  # distribute all handed over params in whole class

    def create_new_study(self) -> optuna.study.Study:
        """
        Create a new optuna study.

        :return: a new optuna study instance
        """
        # outerfold_prefix = \
        #    'OUTER' + self.save_path[[m.end(0) for m in re.finditer(pattern='outerfold_', string=self.save_path)][0]] \
        #    + '-' if 'outerfold' in self.save_path else ''
        outerfold_prefix = \
            'OUTER' + self.save_path.parts[-2].split('_')[1] + '-' if 'outerfold' in self.save_path.parts[-2] else ''
        study_name = datetime.datetime.now().strftime("%Y-%m-%d_%H-%M-%S") + '_' + outerfold_prefix + \
                     self.user_input_params["genotype_matrix_name"].split('.')[0] + '-' + \
                     self.user_input_params["phenotype_matrix_name"].split('.')[0] + '-' + \
                     self.user_input_params["phenotype"] + '-MAF' + str(self.user_input_params["maf_percentage"]) + \
                     '-SPLIT' + self.dataset.datasplit + self.datasplit_subpath + \
                     '-MODEL' + self.current_model_name + '-TRIALS' + str(self.user_input_params["n_trials"])
        # storage = optuna.storages.RDBStorage(
        #     "sqlite:////" + str(self.save_path.joinpath('Optuna_DB.db')), heartbeat_interval=60, grace_period=120,
        #    failed_trial_callback=optuna.storages.RetryFailedTrialCallback(max_retry=3)
        #)
        # TPE Sampler with seed for reproducibility
        # Percentile pruner if minimum 20 trials exist and intermediate result is worse than 80th percentile
        study = optuna.create_study(
            #storage=storage, 
            study_name=study_name,
            direction='minimize' if self.task == 'regression' else 'maximize', load_if_exists=True,
            sampler=optuna.samplers.TPESampler(seed=42),
            pruner=optuna.pruners.PercentilePruner(percentile=80, n_min_trials=20)
        )
        return study

    def objective(self, trial: optuna.trial.Trial, train_val_indices: dict) -> float:
        """
        Objective function for optuna optimization that returns a score

        :param trial: trial of optuna for optimization
        :param train_val_indices: indices of train and validation sets

        :return: score of the current hyperparameter config
        """
        if (trial.number != 0) and (trial.number % self.intermediate_results_interval == 0):
            print('Generate intermediate test results at trial ' + str(trial.number))
            _ = self.generate_results_on_test(outerfold_info=train_val_indices)

        # Setup timers for runtime logging
        start_process_time = time.process_time()
        start_realclock_time = time.time()
        # Create model
        # in case a model has attributes not part of the base class hand them over in a dictionary to keep the same call
        # (name of the attribute and key in the dictionary have to match)
        additional_attributes_dict = {}
        if issubclass(helper_functions.get_mapping_name_to_class()[self.current_model_name],
                      _torch_model.TorchModel) or \
                issubclass(helper_functions.get_mapping_name_to_class()[self.current_model_name],
                           _tensorflow_model.TensorflowModel):
            # additional attributes for torch and tensorflow models
            additional_attributes_dict['n_features'] = self.dataset.X_full.shape[1]
            additional_attributes_dict['batch_size'] = self.user_input_params["batch_size"]
            additional_attributes_dict['n_epochs'] = self.user_input_params["n_epochs"]
            additional_attributes_dict['width_onehot'] = self.dataset.X_full.shape[-1]
            early_stopping_points = []  # log early stopping point at each fold for torch and tensorflow models
        try:
            model: _base_model.BaseModel = helper_functions.get_mapping_name_to_class()[self.current_model_name](
                task=self.task, optuna_trial=trial,
                n_outputs=len(np.unique(self.dataset.y_full)) if self.task == 'classification' else 1,
                **additional_attributes_dict
            )
        except Exception as exc:
            print('Trial failed. Error in model creation.')
            print(exc)
            print(trial.params)
            self.clean_up_after_exception(trial_number=trial.number, trial_params=trial.params,
                                          reason='model creation: ' + str(exc))
            raise optuna.exceptions.TrialPruned()

        # save the unfitted model
        self.save_path.joinpath('temp').mkdir(parents=True, exist_ok=True)
        model.save_model(path=self.save_path.joinpath('temp'),
                         filename='unfitted_model_trial' + str(trial.number))
        print("Params for Trial " + str(trial.number))
        print(trial.params)
        if self.check_params_for_duplicate(current_params=trial.params):
            print('Trial params are a duplicate.')
            self.clean_up_after_exception(trial_number=trial.number, trial_params=trial.params,
                                          reason='pruned: duplicate')
            raise optuna.exceptions.TrialPruned()
        # Iterate over all innerfolds
        objective_values = []
        validation_results = pd.DataFrame(index=range(0, self.dataset.y_full.shape[0]))
        for innerfold_name, innerfold_info in train_val_indices.items():
            # skip test set (see structure described in base_dataset.Dataset)
            if innerfold_name == 'test':
                continue
            if self.dataset.datasplit != 'train-val-test':
                print('# Processing ' + innerfold_name + ' #')
            else:
                innerfold_name = 'train-val'
            # load the unfitted model to prevent information leak between folds
            model = _model_functions.load_model(path=self.save_path.joinpath('temp'),
                                                filename='unfitted_model_trial' + str(trial.number))

            X_train, y_train, sample_ids_train, X_val, y_val, sample_ids_val = \
                self.dataset.X_full[innerfold_info['train']], \
                self.dataset.y_full[innerfold_info['train']], \
                self.dataset.sample_ids_full[innerfold_info['train']], \
                self.dataset.X_full[innerfold_info['val']], \
                self.dataset.y_full[innerfold_info['val']], \
                self.dataset.sample_ids_full[innerfold_info['val']]
            try:
                # run train and validation loop for this fold
                y_pred = model.train_val_loop(X_train=X_train, y_train=y_train, X_val=X_val, y_val=y_val)
                if hasattr(model, 'early_stopping_point'):
                    early_stopping_points.append(
                        model.early_stopping_point if model.early_stopping_point is not None else model.n_epochs)
                if len(y_pred) == (len(y_val) - 1):
                    # might happen if batch size leads to a last batch with only one sample which will be dropped then
                    print('y_val has one element less than y_true (e.g. due to batch size config) -> drop last element')
                    y_val = y_val[:-1]
                objective_value = \
                    sklearn.metrics.accuracy_score(y_true=y_val, y_pred=y_pred) if self.task == 'classification' \
                        else sklearn.metrics.mean_squared_error(y_true=y_val, y_pred=y_pred)
                # report value for pruning
                trial.report(value=objective_value,
                             step=0 if self.dataset.datasplit == 'train-val-test' else int(innerfold_name[-1]))
                if trial.should_prune():
                    self.clean_up_after_exception(trial_number=trial.number, trial_params=trial.params, reason='pruned')
                    raise optuna.exceptions.TrialPruned()
                # store results
                objective_values.append(objective_value)
                validation_results.at[0:len(sample_ids_train) - 1, innerfold_name + '_train_sampleids'] = \
                    sample_ids_train.flatten()
                validation_results.at[0:len(y_train) - 1, innerfold_name + '_train_true'] = y_train.flatten()
                y_train_pred = model.predict(X_in=X_train)
                validation_results.at[0:len(y_train_pred) - 1, innerfold_name + '_train_pred'] = \
                    y_train_pred.flatten()
                validation_results.at[0:len(sample_ids_val) - 1, innerfold_name + '_val_sampleids'] = \
                    sample_ids_val.flatten()
                validation_results.at[0:len(y_val) - 1, innerfold_name + '_val_true'] = y_val.flatten()
                validation_results.at[0:len(y_pred) - 1, innerfold_name + '_val_pred'] = y_pred.flatten()
                for metric, value in eval_metrics.get_evaluation_report(y_pred=y_pred, y_true=y_val, task=self.task,
                                                                        prefix=innerfold_name + '_').items():
                    validation_results.at[0, metric] = value
            except (RuntimeError, TypeError, tf.errors.ResourceExhaustedError, ValueError) as exc:
                print(exc)
                if 'out of memory' in str(exc) or isinstance(exc, tf.errors.ResourceExhaustedError):
                    # Recover from CUDA out of memory error
                    print('CUDA OOM at batch_size ' + str(model.batch_size))
                    del model
                    gc.collect()
                    torch.cuda.empty_cache()
                else:
                    print('Trial failed. Error in optim loop.')
                self.clean_up_after_exception(trial_number=trial.number, trial_params=trial.params,
                                              reason='model optimization: ' + str(exc))
                raise optuna.exceptions.TrialPruned()
        current_val_result = float(np.mean(objective_values))
        if self.current_best_val_result is None or \
                (self.task == 'classification' and current_val_result > self.current_best_val_result) or \
                (self.task == 'regression' and current_val_result < self.current_best_val_result):
            self.current_best_val_result = current_val_result
            if hasattr(model, 'early_stopping_point'):
                # take mean of early stopping points of all innerfolds for refitting of final model
                self.early_stopping_point = int(np.mean(early_stopping_points))
            # persist results
            validation_results.to_csv(
                self.save_path.joinpath('temp', 'validation_results_trial' + str(trial.number) + '.csv'),
                sep=',', decimal='.', float_format='%.10f', index=False
            )
            # delete previous results
            for file in self.save_path.joinpath('temp').iterdir():
                if 'trial' + str(trial.number) not in str(file):
                    self.save_path.joinpath('temp', file).unlink()
        else:
            # delete unfitted model
            self.save_path.joinpath('temp', 'unfitted_model_trial' + str(trial.number)).unlink()
        # save runtime information of this trial
        self.write_runtime_csv(dict_runtime={'Trial': trial.number,
                                             'process_time_s': time.process_time() - start_process_time,
                                             'real_time_s': time.time() - start_realclock_time,
                                             'params': trial.params, 'note': 'successful'})
        return current_val_result

    def clean_up_after_exception(self, trial_number: int, trial_params: dict, reason: str):
        """
        Clean up things after an exception: delete unfitted model if it exists and update runtime csv

        :param trial_number: number of the trial
        :param trial_params: parameters of the trial
        :param reason: hint for the reason of the Exception
        """
        if self.save_path.joinpath('temp', 'unfitted_model_trial' + str(trial_number)).exists():
            self.save_path.joinpath('temp', 'unfitted_model_trial' + str(trial_number)).unlink()
        self.write_runtime_csv(dict_runtime={'Trial': trial_number, 'process_time_s': np.nan, 'real_time_s': np.nan,
                                             'params': trial_params, 'note': reason})

    def write_runtime_csv(self, dict_runtime: dict):
        """
        Write runtime info to runtime csv file

        :param dict_runtime: dictionary with runtime information
        """
        with open(self.save_path.joinpath(self.current_model_name + '_runtime_overview.csv'), 'a') as runtime_file:
            headers = ['Trial', 'process_time_s', 'real_time_s', 'params', 'note']
            writer = csv.DictWriter(f=runtime_file, fieldnames=headers)
            if runtime_file.tell() == 0:
                writer.writeheader()
            writer.writerow(dict_runtime)

    def calc_runtime_stats(self) -> dict:
        """
        Calculate runtime stats for saved csv file.

        :return: dict with runtime info enhanced with runtime stats
        """
        csv_file = pd.read_csv(self.save_path.joinpath(self.current_model_name + '_runtime_overview.csv'))
        if csv_file['Trial'].dtype is object and any(["retrain" in elem for elem in csv_file["Trial"]]):
            csv_file = csv_file[csv_file["Trial"].str.contains("retrain") == False]
        process_times = csv_file['process_time_s']
        real_times = csv_file['real_time_s']
        process_time_mean, process_time_std, process_time_max, process_time_min = \
            process_times.mean(), process_times.std(), process_times.max(), process_times.min()
        real_time_mean, real_time_std, real_time_max, real_time_min = \
            real_times.mean(), real_times.std(), real_times.max(), real_times.min()
        self.write_runtime_csv({'Trial': 'mean', 'process_time_s': process_time_mean, 'real_time_s': real_time_mean})
        self.write_runtime_csv({'Trial': 'std', 'process_time_s': process_time_std, 'real_time_s': real_time_std})
        self.write_runtime_csv({'Trial': 'max', 'process_time_s': process_time_max, 'real_time_s': real_time_max})
        self.write_runtime_csv({'Trial': 'min', 'process_time_s': process_time_min, 'real_time_s': real_time_min})
        return {'process_time_mean': process_time_mean, 'process_time_std': process_time_std,
                'process_time_max': process_time_max, 'process_time_min': process_time_min,
                'real_time_mean': real_time_mean, 'real_time_std': real_time_std,
                'real_time_max': real_time_max, 'real_time_min': real_time_min}

    def check_params_for_duplicate(self, current_params: dict) -> bool:
        """
        Check if params were already suggested which might happen by design of TPE sampler.

        :param current_params: dictionar with current parameters

        :return: bool reflecting if current params were already used in the same study
        """
        past_params = [trial.params for trial in self.study.trials[:-1]]
        return current_params in past_params

    def generate_results_on_test(self, outerfold_info: dict) -> dict:
        """
        Generate the results on the testing data

        :param outerfold_info: dictionary with outerfold datasplit indices

        :return: evaluation metrics dictionary
        """

        helper_functions.set_all_seeds()
        # Retrain on full train + val data with best hyperparams and apply on test
        print("## Retrain best model and test ##")
        X_test, y_test, sample_ids_test = \
            self.dataset.X_full[outerfold_info['test']], self.dataset.y_full[outerfold_info['test']], \
            self.dataset.sample_ids_full[outerfold_info['test']]
        X_retrain, y_retrain, sample_ids_retrain = \
            self.dataset.X_full[~np.isin(np.arange(len(self.dataset.X_full)), outerfold_info['test'])], \
            self.dataset.y_full[~np.isin(np.arange(len(self.dataset.y_full)), outerfold_info['test'])], \
            self.dataset.sample_ids_full[~np.isin(np.arange(len(self.dataset.sample_ids_full)),
                                                  outerfold_info['test'])]
        start_process_time = time.process_time()
        start_realclock_time = time.time()
        postfix = '' if len(self.study.trials) == self.user_input_params["n_trials"] else 'temp'
        final_model = _model_functions.load_retrain_model(
            path=self.save_path.joinpath(postfix), filename='unfitted_model_trial' + str(self.study.best_trial.number),
            X_retrain=X_retrain, y_retrain=y_retrain, early_stopping_point=self.early_stopping_point)
        y_pred_retrain = final_model.predict(X_in=X_retrain)
        no_trials = len(self.study.trials) - 1 if len(self.study.trials) % self.intermediate_results_interval != 0 \
            else len(self.study.trials)
        self.write_runtime_csv(dict_runtime={'Trial': 'retraining_after_' + str(no_trials),
                                             'process_time_s': time.process_time() - start_process_time,
                                             'real_time_s': time.time() - start_realclock_time,
                                             'params': self.study.best_trial.params, 'note': 'successful'})
        y_pred_test = final_model.predict(X_in=X_test)

        # Evaluate and save results
        eval_scores = \
            eval_metrics.get_evaluation_report(y_pred=y_pred_test, y_true=y_test, task=self.task, prefix='test_')

        feat_import_df = None
        if self.current_model_name in ['randomforest', 'xgboost', 'linearregression', 'elasticnet']:
            feat_import_df = self.get_feature_importance(model=final_model, X=X_test, y=y_test)

        print('## Results on test set ##')
        print(eval_scores)
        final_results = pd.DataFrame(index=range(0, self.dataset.y_full.shape[0]))
        final_results.at[0:len(sample_ids_retrain) - 1, 'sample_ids_retrain'] = sample_ids_retrain.flatten()
        final_results.at[0:len(y_pred_retrain) - 1, 'y_pred_retrain'] = y_pred_retrain.flatten()
        final_results.at[0:len(y_retrain) - 1, 'y_true_retrain'] = y_retrain.flatten()
        final_results.at[0:len(sample_ids_test) - 1, 'sample_ids_test'] = sample_ids_test.flatten()
        final_results.at[0:len(y_pred_test) - 1, 'y_pred_test'] = y_pred_test.flatten()
        final_results.at[0:len(y_test) - 1, 'y_true_test'] = y_test.flatten()
        for metric, value in eval_scores.items():
            final_results.at[0, metric] = value
        if len(self.study.trials) == self.user_input_params["n_trials"]:
            results_filename = 'final_model_test_results.csv'
            feat_import_filename = 'final_model_feature_importances.csv'
            if self.user_input_params["save_final_model"]:
                final_model.save_model(path=self.save_path, filename='final_retrained_model')
        else:
            results_filename = 'intermediate_after_' + str(len(self.study.trials) - 1) + '_test_results.csv'
            feat_import_filename = \
                'intermediate_after_' + str(len(self.study.trials) - 1) + '_feat_importances.csv'
            shutil.copyfile(self.save_path.joinpath(self.current_model_name + '_runtime_overview.csv'),
                            self.save_path.joinpath('temp',
                                                    'intermediate_after_' + str(len(self.study.trials) - 1) + '_' +
                                                    self.current_model_name + '_runtime_overview.csv'), )
        final_results.to_csv(
            self.save_path.joinpath(postfix, results_filename),
            sep=',', decimal='.', float_format='%.10f', index=False
        )
        if feat_import_df is not None:
            feat_import_df.to_csv(
                self.save_path.joinpath(postfix, feat_import_filename),
                sep=',', decimal='.', float_format='%.10f', index=False
            )
        return eval_scores

    def get_feature_importance(self, model: _base_model.BaseModel, X: np.array, y: np.array,
                               top_n: int = 1000, include_perm_importance: bool = False) -> pd.DataFrame:
        """
        Get feature importances for models that possess such a feature, e.g. XGBoost

        :param model: model to analyze
        :param X: feature matrix for permutation
        :param y: target vector for permutation
        :param top_n: top n features to select
        :param include_perm_importance: include permutation based feature importance or not

        :return: DataFrame with feature importance information
        """

        top_n = min(len(self.dataset.snp_ids), top_n)
        feat_import_df = pd.DataFrame()
        if self.current_model_name in ['randomforest', 'xgboost']:
            feature_importances = model.model.feature_importances_
            sorted_idx = feature_importances.argsort()[::-1][:top_n]
            feat_import_df['snp_ids_standard'] = self.dataset.snp_ids[sorted_idx]
            feat_import_df['feat_importance_standard'] = feature_importances[sorted_idx]
        else:
            coefs = model.model.coef_
            dims = coefs.shape[0] if len(coefs.shape) > 1 else 1
            for dim in range(dims):
                coef = coefs[dim] if len(coefs.shape) > 1 else coefs
                sorted_idx = coef.argsort()[::-1][:top_n]
                feat_import_df['snp_ids_' + str(dim)] = self.dataset.snp_ids[sorted_idx]
                feat_import_df['coefficients_' + str(dim)] = coef[sorted_idx]
        if include_perm_importance:
            perm_importance = sklearn.inspection.permutation_importance(
                estimator=model.model, X=X, y=y
            )
            sorted_idx = perm_importance.importances_mean.argsort()[::-1][:top_n]
            feat_import_df['snp_ids_perm'] = self.dataset.snp_ids[sorted_idx]
            feat_import_df['feat_importance_perm_mean'] = perm_importance.importances_mean[sorted_idx]
            feat_import_df['feat_importance_perm_std'] = perm_importance.importances_std[sorted_idx]

        return feat_import_df

    def run_optuna_optimization(self) -> dict:
        """
        Run whole optuna optimization for one model, dataset and datasplit.

        :return: dictionary with results overview
        """
        # Iterate over outerfolds
        # (according to structure described in base_dataset.Dataset, only for nested-cv multiple outerfolds exist)
        helper_functions.set_all_seeds()
        devices = tf.config.list_physical_devices('GPU')
        if devices:
            try:
                for gpu in devices:
                    tf.config.experimental.set_memory_growth(gpu, True)
            except RuntimeError as e:
                print(e)
        overall_results = {}
        for outerfold_name, outerfold_info in self.dataset.datasplit_indices.items():
            if self.dataset.datasplit == 'nested-cv':
                if self.user_input_params["outerfold_number_to_run"] is not None and \
                        outerfold_name != 'outerfold_' + str(self.user_input_params["outerfold_number_to_run"]):
                    continue
                # Only print outerfold info for nested-cv as it does not apply for the other splits
                print("## Starting Optimization for " + outerfold_name + " ##")
                #end_ind = [m.end(0) for m in re.finditer(pattern='/', string=self.base_path)][-2]
                #self.save_path = self.base_path[:end_ind] + outerfold_name + '/' + self.base_path[end_ind:]
                self.save_path = self.base_path.parent.joinpath(outerfold_name, self.base_path.name)
            if not self.save_path.exists():
                self.save_path.mkdir(parents=True, exist_ok=True)
            # Create a new study for each outerfold
            self.study = self.create_new_study()
            self.current_best_val_result = None
            # Start optimization run
            self.study.optimize(
                lambda trial: self.objective(trial=trial, train_val_indices=outerfold_info),
                n_trials=self.user_input_params["n_trials"]
            )
            helper_functions.set_all_seeds()
            # set seeds again after optuna runs are finished as number of trials might lead to different states of
            # random number generators
            # Calculate runtime metrics after finishing optimization
            runtime_metrics = self.calc_runtime_stats()
            # Print statistics after run
            print("## Optuna Study finished ##")
            print("Study statistics: ")
            print("  Finished trials: ", len(self.study.trials))
            print("  Pruned trials: ", len(self.study.get_trials(states=(optuna.trial.TrialState.PRUNED,))))
            print("  Completed trials: ", len(self.study.get_trials(states=(optuna.trial.TrialState.COMPLETE,))))
            print("  Best Trial: ", self.study.best_trial.number)
            print("  Value: ", self.study.best_trial.value)
            print("  Params: ")
            for key, value in self.study.best_trial.params.items():
                print("    {}: {}".format(key, value))

            # Move validation results and models of best trial
            # files_to_keep = glob.glob(self.save_path + 'temp/' + '*trial' + str(self.study.best_trial.number) + '*')
            files_to_keep_path = self.save_path.joinpath('temp', '*trial' + str(self.study.best_trial.number) + '*')
            files_to_keep = pathlib.Path(files_to_keep_path.parent).expanduser().glob(files_to_keep_path.name)
            for file in files_to_keep:
                shutil.copyfile(file, self.save_path.joinpath(file.name))
            shutil.rmtree(self.save_path.joinpath('temp'))

            # Retrain on full train + val data with best hyperparams and apply on test
            eval_scores = self.generate_results_on_test(outerfold_info=outerfold_info)
            key = outerfold_name if self.dataset.datasplit == 'nested-cv' else 'Test'
            best_params = self.study.best_trial.params
            if issubclass(helper_functions.get_mapping_name_to_class()[self.current_model_name],
                          _torch_model.TorchModel) or \
                    issubclass(helper_functions.get_mapping_name_to_class()[self.current_model_name],
                               _tensorflow_model.TensorflowModel):
                # additional attributes for torch and tensorflow models
                if 'n_epochs' not in best_params.keys():
                    best_params['n_epochs'] = self.user_input_params["n_epochs"]
                if 'batch_size' not in best_params.keys():
                    best_params['batch_size'] = self.user_input_params["batch_size"]
                best_params['early_stopping_point'] = self.early_stopping_point

            overall_results[key] = {'best_params': best_params, 'eval_metrics': eval_scores,
                                    'runtime_metrics': runtime_metrics}
        return overall_results



import numpy as np
import pandas as pd
import pathlib
import random

from ..preprocess import raw_data_functions, encoding_functions
from ..utils import check_functions


def filter_duplicates(X: np.array, snp_ids: np.array) -> (np.array, np.array):
    """
    Remove duplicate SNPs, i.e. SNPs that are completely the same for all samples and therefore do not add information.

    :param X: genotype matrix to be filtered
    :param snp_ids: vector containing corresponding SNP ids

    :return: filtered genotype matrix and filtered SNP ids
    """
    uniques, index = np.unique(X, return_index=True, axis=1)
    X = uniques[:, np.argsort(index)]
    snp_ids = snp_ids[np.sort(index)]
    return X, snp_ids


def get_simulation(X: np.array, sample_ids: np.array, snp_ids: np.array, number_of_samples: int, number_causal_snps: int,
                   explained_variance: int, maf: int, heritability: int, seed: int,
                   number_background_snps: int, distribution: str, shape: float) \
                    -> (np.array, np.array, np.array, np.array, np.array, np.array, np.array):
    """
    Simulate phenotypes based on (real) genotypes in an additive setting with normally or gamma distributed noise and
    normally distributed effect sizes of causal SNPs.

    :param X: genotype matrix in additive encoding
    :param sample_ids: sample ids of genotype matrix
    :param snp_ids: SNP ids of genotype matrix
    :param number_of_samples: number of samples of synthetic phenotype
    :param number_causal_snps: number of SNPs used as causal markers in simulation
    :param explained_variance: percentage value of how much of the total variance the causal SNPs should explain
    :param maf: percentage value used for maf filtering of genotype matrix
    :param heritability: percentage value of how much of the variance should be explained by polygenic background
    :param seed: seed for random sampling
    :param number_background_snps: number of randomly selected SNPs to simulate the polygenic background
    :param distribution: probability distribution used to draw random noise can be 'normal' or 'gamma'
    :param shape: only needed if distribution is 'gamma'

    :return: simulated phenotype with corresponding sample ids, SNP ids of causal SNPs, SNP ids of background SNPs,
    effect sizes of background, effect sizes of causal SNPs, used explained variance for each causal SNP
    """

    # sanity checks
    if encoding_functions.check_encoding_of_genotype(X=X) != '012':
        raise Exception('Encoding of genotype must be additive!')
    if distribution == 'gamma' and shape is None:
        raise Exception('Need shape value to use gamma distribution!')
    if number_background_snps < 1:
        raise Exception('Need at least one background SNP.')
    if number_of_samples < 1:
        raise Exception('Need at least one sample.')
    if number_of_samples > len(sample_ids):
        print('Only %d samples are available. Cannot choose %d samples for simulations. Will use all available samples '
              'instead.' % (len(sample_ids), number_of_samples))
        number_of_samples = len(sample_ids)

    # get random samples for X
    random.seed(seed)
    np.random.seed(seed)
    samples_to_take = random.sample(list(enumerate(sample_ids)), number_of_samples)
    sample_indices = np.array(samples_to_take)[:, 0].astype(int)
    sample_ids_sampled = np.array(samples_to_take)[:, 1]
    X_sampled = X[sample_indices, :]

    # filter non-informative
    X_sampled, X_index = raw_data_functions.filter_non_informative_snps(X=X_sampled)
    snp_ids_sampled = snp_ids[X_index]

    # filter for duplicates
    X_sampled, snp_ids_sampled = filter_duplicates(X_sampled, snp_ids_sampled)

    # filter for MAF
    freq = raw_data_functions.get_minor_allele_freq(X=X_sampled)
    filter_indices = raw_data_functions.create_maf_filter(maf=maf, freq=freq)
    X_sampled = np.delete(X_sampled, filter_indices, axis=1)
    snp_ids_sampled = np.delete(snp_ids_sampled, filter_indices)

    # sanity checks
    if number_causal_snps > len(snp_ids_sampled):
        raise Exception('After filtering only %d SNPs remain. Not enough SNPs available to simulate %d causal SNPs.'
                        % (len(snp_ids_sampled), number_causal_snps))
    if number_causal_snps + number_background_snps > len(snp_ids_sampled):
        raise Exception('Only %d SNPs are available after filtering. Cannot choose %d causal SNPs and %d background SNPs'
                        ' for simulations. Please check again'
                        % (len(snp_ids_sampled), number_causal_snps, number_background_snps))

    # compute simulations
    # choose random causal SNPs
    if number_causal_snps >= 1:
        causal_snps = random.sample(list(enumerate(snp_ids_sampled)), number_causal_snps)
        causal_snps_indices = np.array(causal_snps)[:, 0].astype(int)
        causal_snps_ids = np.array(causal_snps)[:, 1]
    else:
        causal_snps_indices = []
        causal_snps_ids = []

    # choose background SNPs
    X_non_causal = np.delete(X_sampled, causal_snps_indices, axis=1)
    snp_ids_non_causal = np.delete(snp_ids_sampled, causal_snps_indices, axis=0)
    background_SNPs_indices = np.random.choice(X_non_causal.shape[1], number_background_snps, replace=False)
    background_SNPs = X_non_causal[:, background_SNPs_indices]
    background_snp_ids = snp_ids_non_causal[background_SNPs_indices]

    # compute effect size for background
    betas_background = np.random.normal(loc=0, scale=0.1, size=number_background_snps)

    # add background
    simulated_phenotype = np.matmul(background_SNPs, betas_background)

    # set heritability
    heritability = heritability / 100
    background_variance = np.var(simulated_phenotype)
    noise_variance = background_variance / heritability - background_variance

    # add random noise
    if distribution == 'gamma':
        random_noise = np.random.gamma(shape=shape, scale=np.sqrt(noise_variance / shape), size=number_of_samples)
    elif distribution == 'normal':
        random_noise = np.random.normal(loc=0, scale=np.sqrt(noise_variance), size=number_of_samples)
    else:
        raise Exception("Can only simulate noise with 'gamma' or 'normal' distribution.")
    simulated_phenotype = simulated_phenotype + random_noise

    # compute explained variances for more than 1 snp
    if number_causal_snps < 1:
        return simulated_phenotype, sample_ids_sampled, causal_snps_ids, background_snp_ids, betas_background, [], []
    if number_causal_snps > 1:
        c = explained_variance
        mean = c / number_causal_snps
        sd = (mean / c) * 10
        explained_variance = np.random.normal(mean, sd, number_causal_snps)
        explained_variance.sort()
    else:
        explained_variance = np.array([explained_variance])
    explained_variance = explained_variance / 100

    # add causative markers with effect sizes
    caus_beta = []
    for i in range(number_causal_snps):
        beta = np.sqrt((explained_variance[i] / (1 - explained_variance[i]) *
                        (np.var(simulated_phenotype) / np.var(X_sampled[:, causal_snps_indices[i]]))))
        simulated_phenotype += beta * X_sampled[:, causal_snps_indices[i]]
        caus_beta.append(beta)

    return simulated_phenotype, sample_ids_sampled, causal_snps_ids, background_snp_ids, betas_background, caus_beta, \
           explained_variance


def check_sim_id(sim_dir: pathlib.Path) -> int:
    """
    Check which ids were already used for simulations.

    :param sim_dir: directory containing simulations to check

    :return: last simulation number + 1
    """
    sim_ids = []
    for sim in sim_dir.iterdir():
        if sim.is_file() and 'Simulation_' in sim.as_posix():
            sim_numbers = sim.with_suffix('').name.split('_')[-1]
            sim_ids.append(int(sim_numbers.split('-')[-1]))
    if len(sim_ids) == 0:
        return 1
    else:
        return max(sim_ids) + 1


def save_sim_overview(save_dir: pathlib.Path, sim_names: list, number_of_samples: list, number_causal_snps: list,
                      explained_variance: list, maf: list, heritability: list, seeds: list,
                      number_background_snps: list, distribution: list, shape: list):
    """
    save overview file for all simulations; append new simulations if file already exists

    :param save_dir: directory to save overview file to
    :param sim_names: list containing simulation name for each simulation
    :param number_of_samples: list containing number of samples for each simulation
    :param number_causal_snps: list containing number of causal SNPS for each simulation
    :param explained_variance: list containing total explained variance of causal SNPs for each simulation
    :param maf: list containing used maf frequency for each simulation
    :param heritability: list containing used heritability for each simulation
    :param seeds: list containing used seed for each simulation
    :param number_background_snps: list containing number of background SNPs for each simulation
    :param distribution: list containing used distribution of random noise for each simulation
    :param shape: list containing shape of gamma distribution, resp. None for normal distribution for each simulation
    """
    overview_file = save_dir.joinpath('Simulations_Overview.csv')
    if not check_functions.check_exist_files([overview_file]):
        print('Create new overview file for simulations.')
        df_sim = pd.DataFrame({'simulation': sim_names,
                               'seed': seeds,
                               'heritability': heritability,
                               'MAF': maf,
                               'samples': number_of_samples,
                               'causal_SNPs': number_causal_snps,
                               'background_SNPs': number_background_snps,
                               'explained_var': explained_variance,
                               'distribution': distribution,
                               'shape': shape})
    else:
        print('Overview file already exists. Will append new simulations to file.')
        df_old = pd.read_csv(overview_file)
        df_new = pd.DataFrame({'simulation': sim_names,
                               'seed': seeds,
                               'heritability': heritability,
                               'MAF': maf,
                               'samples': number_of_samples,
                               'causal_SNPs': number_causal_snps,
                               'background_SNPs': number_background_snps,
                               'explained_var': explained_variance,
                               'distribution': distribution,
                               'shape': shape})
        df_sim = pd.concat([df_old, df_new])
    df_sim.to_csv(overview_file, index=False)


def save_simulation(save_dir: str, genotype_matrix_name: str, number_of_sim: int, X: np.array, sample_ids: np.array,
                    snp_ids: np.array, number_of_samples: int, number_causal_snps: int, explained_variance: int,
                    maf: int, heritability: int, seed: int, number_background_snps: int, distribution: str,
                    shape: float):
    """
    Set all variables and generate one or more simulations with same configurations.
    Save overview file and simulated phenotypes to subfolder 'genotype_matrix_name' in save_dir as
    'Simulations_Overview.csv' and Simulation_{sim_id}.csv.
    Save SNP ids of background SNPs, effect sizes/betas of background SNPs and configuration infos containing SNP ids
    and betas of causal SNPs to subfolder sim_configs in 'genotype_matrix_name' as 'background_{sim_id}.csv',
    'betas_background_{sim_id}.csv' and 'simulation_config_{sim_id}.csv'.
    If only one phenoype is simulated, the sim_id consists of a single number. If several phenotypes are simulated with
    the same configurations, then the sim_id is the number of the first simulation '-' number of last simulation,
    e.g. '10-15'

    :param save_dir: directory to save simulations to
    :param genotype_matrix_name: name of genotype matrix to be used for simulations, needed to create subfolder in save_dir
    :param number_of_sim: number of simulations to create with same configurations
    :param X: genotype matrix in additive encoding
    :param sample_ids: sample ids of genotype matrix
    :param snp_ids: SNP ids of genotype matrix
    :param number_of_samples: number of samples of synthetic phenotype
    :param number_causal_snps: number of SNPs used as causal markers in simulation
    :param explained_variance: percentage value of how much of the total variance the causal SNPs should explain
    :param maf: percentage value used for maf filtering of genotype matrix
    :param heritability: percentage value of how much of the variance should be explained by polygenic background
    :param seed: seed for random sampling
    :param number_background_snps: number of randomly selected SNPs to simulate the polygenic background
    :param distribution: probability distribution used to draw random noise can be 'normal' or 'gamma'
    :param shape: only needed if distribution is 'gamma'
    """

    # check directories
    save_dir = pathlib.Path(save_dir).joinpath(genotype_matrix_name).with_suffix('')
    check_functions.check_exist_directories(list_of_dirs=[save_dir, save_dir.joinpath('sim_configs')],
                                            create_if_not_exist=True)

    # fix numbers / names of simulations
    sim_number = int(check_sim_id(sim_dir=save_dir))
    if number_of_sim > 1:
        sim_names = np.arange(sim_number, sim_number + number_of_sim)
        sim_id = str(sim_names[0]) + '-' + str(sim_names[-1])
    elif number_of_sim == 1:
        sim_names = [sim_number]
        sim_id = str(sim_number)
    else:
        raise Exception('number of simulations has to be at least 1')

    print('Now create %d simulations with %d samples, %d causal SNPs, %d background SNPs, '
          'heritability of %d, %d explained variance, %d maf and %s distribution'
          % (number_of_sim, number_of_samples, number_causal_snps, number_background_snps, heritability,
            explained_variance, maf, distribution))
    print('Save simulations with sim_id ' + sim_id)

    # create simulations
    causal_markers = []
    seeds = []
    background_markers = []
    background_betas = []
    causative_beta = []
    ev = []
    df_final = pd.DataFrame(index=sample_ids)
    for i in range(number_of_sim):
        new_seed = seed + sim_names[i]
        simulated_phenotype, sample_ids_sampled, causal_snps_ids, background_snp_ids, betas_background, beta, c = \
            get_simulation(X, sample_ids, snp_ids, number_of_samples, number_causal_snps, explained_variance, maf,
                           heritability, new_seed, number_background_snps, distribution, shape)

        causal_markers.append(causal_snps_ids)
        seeds.append(new_seed)
        background_markers.append(background_snp_ids)
        background_betas.append(betas_background)
        causative_beta.append(beta)
        ev.append(c)

        df_sim = pd.DataFrame({f'sim{sim_names[i]}': simulated_phenotype,
                               f'sim{sim_names[i]}_shift': simulated_phenotype + np.abs(
                                   np.min(simulated_phenotype)) + 1},
                              index=sample_ids_sampled)
        df_final = df_final.join(df_sim)

    # save overview
    save_sim_overview(save_dir=save_dir, sim_names=sim_names, number_of_samples=[number_of_samples] * number_of_sim,
                      number_causal_snps=[number_causal_snps] * number_of_sim,
                      explained_variance=[explained_variance] * number_of_sim, maf=[maf] * number_of_sim,
                      heritability=[heritability] * number_of_sim, seeds=seeds,
                      number_background_snps=[number_background_snps] * number_of_sim,
                      distribution=[distribution] * number_of_sim, shape=[shape] * number_of_sim)

    # save simulations
    df_final.to_csv(save_dir.joinpath(f'Simulation_{sim_id}.csv'))

    # save configs
    df_causal = pd.DataFrame({'simulation': sim_names,
                              'seed': seeds,
                              'heritability': heritability,
                              'samples': number_of_samples,
                              'SNPs': number_causal_snps,
                              'explained_var': ev,
                              'causal_marker': causal_markers,
                              'causal_beta': causative_beta,
                              'distribution': distribution,
                              'shape': shape})

    df_causal.to_csv(save_dir.joinpath('sim_configs', f'simulation_config_{sim_id}.csv'), index=False)

    # save background markers and betas
    col = []
    for sim in sim_names:
        col.append('sim' + str(sim))
    col = tuple(col)
    bg = np.array(background_markers).T
    df_background = pd.DataFrame(bg, columns=col)
    df_background.to_csv(save_dir.joinpath('sim_configs', f'background_{sim_id}.csv'), index=False)
    bb = np.array(background_betas).T
    df_bb = pd.DataFrame(bb, columns=col)
    df_bb.to_csv(save_dir.joinpath('sim_configs', f'betas_background_{sim_id}.csv'), index=False)



import numpy as np
import h5py
import pandas as pd
import pathlib
import datetime

from ..preprocess import base_dataset, raw_data_functions
from ..utils import helper_functions, check_functions
from ..model import _model_functions, _torch_model, _tensorflow_model, _base_model
from ..evaluation import eval_metrics
from . import feat_importance, results_analysis


def apply_final_model(results_directory_model: str, old_data_dir: str, new_data_dir: str,
                      new_genotype_matrix: str, new_phenotype_matrix: str, save_dir: str = None):
    """
    Apply a final model on a new dataset. It will be applied to the whole dataset.
    So the main purpose of this function is, if you get new samples you want to predict on.
    If the final model was saved, this will be used for inference on the new dataset.
    Otherwise, it will be retrained on the initial dataset and then used for inference on the new dataset.

    The new dataset will be filtered for the SNP ids that the model was initially trained on.

    CAUTION: the SNPs of the old and the new dataset have to be the same!

    :param results_directory_model: directory that contains the model results that you want to use
    :param old_data_dir: directory that contains the data that the model was trained on
    :param new_data_dir: directory that contains the new genotype and phenotype matrix
    :param new_genotype_matrix: new genotype matrix (incl. file suffix)
    :param new_phenotype_matrix: new phenotype matrix (incl. file suffix)
    :param save_dir: directory to store the results
    """
    results_directory_model = pathlib.Path(results_directory_model)
    old_data_dir = pathlib.Path(old_data_dir)
    new_data_dir = pathlib.Path(new_data_dir)
    save_dir = pathlib.Path(save_dir)

    # Check user inputs
    print("Checking user inputs")
    if not check_functions.check_exist_directories(
            list_of_dirs=[results_directory_model, new_data_dir, save_dir, old_data_dir]):
        raise Exception("See output above. Problems with specified directories")

    result_folder_name = results_directory_model.parts[-3] if 'nested' in str(results_directory_model) \
        else results_directory_model.parts[-2]
    model_name = results_directory_model.parts[-1]
    if 'fromR' in model_name:
        import easypheno.model
        easypheno.model.__all__.extend(['_bayesfromR', 'bayesAfromR', 'bayesBfromR', 'bayesCfromR'])
    datasplit, n_outerfolds, n_innerfolds, val_set_size_percentage, test_set_size_percentage, maf_perc = \
        helper_functions.get_datasplit_config_info_for_resultfolder(resultfolder=result_folder_name)
    nested_offset = -1 if 'nested' in datasplit else 0
    phenotype = results_directory_model.parts[-3 + nested_offset]
    old_genotype_matrix = results_directory_model.parts[-5 + nested_offset]
    old_phenotype_matrix = results_directory_model.parts[-4 + nested_offset]
    if not check_functions.check_exist_files(
            list_of_files=[
                new_data_dir.joinpath(new_phenotype_matrix), new_data_dir.joinpath(new_genotype_matrix),
                old_data_dir.joinpath(base_dataset.Dataset.get_index_file_name(
                    genotype_matrix_name=old_genotype_matrix, phenotype_matrix_name=old_phenotype_matrix,
                    phenotype=phenotype))]):
        raise Exception("See output above. Problems with specified files.")

    # Prepare the new data
    print("Preparing the new dataset")
    raw_data_functions.prepare_data_files(
        data_dir=new_data_dir, genotype_matrix_name=new_genotype_matrix, phenotype_matrix_name=new_phenotype_matrix,
        phenotype=phenotype, datasplit=datasplit, n_outerfolds=n_outerfolds, n_innerfolds=n_innerfolds,
        test_set_size_percentage=test_set_size_percentage, val_set_size_percentage=val_set_size_percentage,
        models=[model_name], user_encoding=None, maf_percentage=maf_perc
    )
    encoding = helper_functions.get_mapping_name_to_class()[model_name].standard_encoding
    new_dataset = base_dataset.Dataset(
        data_dir=new_data_dir, genotype_matrix_name=new_genotype_matrix, phenotype_matrix_name=new_phenotype_matrix,
        phenotype=phenotype, datasplit=datasplit, n_outerfolds=n_outerfolds, n_innerfolds=n_innerfolds,
        test_set_size_percentage=test_set_size_percentage, val_set_size_percentage=val_set_size_percentage,
        encoding=encoding, maf_percentage=maf_perc, do_snp_filters=False
    )

    # Check and filter SNPids in comparison with old dataset
    with h5py.File(old_data_dir.joinpath(
            base_dataset.Dataset.get_index_file_name(genotype_matrix_name=old_genotype_matrix,
                                                     phenotype_matrix_name=old_phenotype_matrix,
                                                     phenotype=phenotype)), "r") as f:
        old_dataset_snp_ids = f[f'matched_data/final_snp_ids/{encoding}/maf_{maf_perc}_snp_ids'][:].astype(str)
    if not check_functions.compare_snp_id_vectors(snp_id_vector_big_equal=new_dataset.snp_ids,
                                                  snp_id_vector_small_equal=old_dataset_snp_ids):
        raise Exception('SNPids of initial dataset and new dataset do not match.')
    old_dataset_snp_ids = np.asarray(old_dataset_snp_ids, dtype=new_dataset.snp_ids.dtype).flatten()
    _, ids_to_keep = \
        (np.reshape(old_dataset_snp_ids, (old_dataset_snp_ids.shape[0], 1)) == new_dataset.snp_ids).nonzero()
    new_dataset.X_full = new_dataset.X_full[:, ids_to_keep]

    # Prepare the model
    outerfold_number = int(results_directory_model.parent.parts[-1].split('_')[1]) if 'nested' in datasplit else 0
    models = results_directory_model.parts[-2 + nested_offset].split('_')[3].split('+')
    results_file_path = \
        results_directory_model.parents[0 - nested_offset].joinpath('Results_overview_' + '_'.join(models) + '.csv')
    full_model_path = results_directory_model.joinpath("final_retrained_model")
    if full_model_path.is_file():
        print("Loading saved model")
        model = _model_functions.load_model(path=results_directory_model, filename=full_model_path.parts[-1])
    else:
        print("Retraining model")
        print("Loading old dataset")
        old_dataset = base_dataset.Dataset(
            data_dir=old_data_dir, genotype_matrix_name=old_genotype_matrix, phenotype_matrix_name=old_phenotype_matrix,
            phenotype=phenotype, datasplit=datasplit, n_outerfolds=n_outerfolds, n_innerfolds=n_innerfolds,
            test_set_size_percentage=test_set_size_percentage, val_set_size_percentage=val_set_size_percentage,
            encoding=encoding, maf_percentage=maf_perc
        )
        if not results_file_path.is_file():
            raise Exception("Results Overview file not existing. Please check: " + str(results_file_path))
        model = _model_functions.retrain_model_with_results_file(
            results_file_path=results_file_path, model_name=model_name, datasplit=datasplit,
            outerfold_number=outerfold_number, dataset=old_dataset
        )

    # Do inference and save results
    print('-----------------------------------------------')
    print("Inference on new data for " + model_name)
    y_pred_new_dataset = model.predict(X_in=new_dataset.X_full)
    eval_scores = \
        eval_metrics.get_evaluation_report(y_pred=y_pred_new_dataset, y_true=new_dataset.y_full, task=model.task,
                                           prefix='test_')
    print('New dataset: ')
    print(eval_scores)
    if results_file_path.is_file():
        print('Old dataset: ')
        results = pd.read_csv(results_file_path)
        results = results[results[results.columns[0]] == 'outerfold_' + str(outerfold_number)] \
            if datasplit == 'nested-cv' else results
        results = results.loc[:, [model_name in col for col in results.columns]]
        eval_dict_saved = results_analysis.result_string_to_dictionary(
            result_string=results[model_name + '___eval_metrics'][outerfold_number]
        )
        print(eval_dict_saved)
    final_results = pd.DataFrame(index=range(0, new_dataset.y_full.shape[0]))
    final_results.at[0:len(new_dataset.sample_ids_full) - 1, 'sample_ids'] = new_dataset.sample_ids_full.flatten()
    final_results.at[0:len(y_pred_new_dataset) - 1, 'y_pred_test'] = y_pred_new_dataset.flatten()
    final_results.at[0:len(new_dataset.y_full) - 1, 'y_true_test'] = new_dataset.y_full.flatten()
    for metric, value in eval_scores.items():
        final_results.at[0, metric] = value
    final_results.at[0, 'base_model_path'] = results_directory_model
    models_start_time = model_name + '_' + datetime.datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
    final_results.to_csv(save_dir.joinpath(
        'predict_results_on-' + new_dataset.index_file_name.split('.')[0] + '-' + models_start_time + '.csv'),
        sep=',', decimal='.', float_format='%.10f', index=False
    )


def retrain_on_new_data(results_directory_model: str, data_dir: str, genotype_matrix: str, phenotype_matrix: str,
                        phenotype: str, encoding: str = None, maf_percentage: int = 0, save_dir: str = None,
                        datasplit: str = 'nested-cv', n_outerfolds: int = 5, n_innerfolds: int = 5,
                        test_set_size_percentage: int = 20, val_set_size_percentage: int = 20,
                        save_final_model: bool = True):
    """
    Train a model on a new dataset using the hyperparameters that worked best for the specified model results.

    :param data_dir: data directory where the phenotype and genotype matrix are stored
    :param genotype_matrix: name of the genotype matrix including datatype ending
    :param phenotype_matrix: name of the phenotype matrix including datatype ending
    :param phenotype: name of the phenotype to predict
    :param encoding: encoding to use. Default is None, so standard encoding of each model will be used. Options are: '012', 'onehot', 'raw'
    :param maf_percentage: threshold for MAF filter as percentage value. Default is 0, so no MAF filtering
    :param save_dir: directory for saving the results. Default is None, so same directory as data_dir
    :param datasplit: datasplit to use. Options are: nested-cv, cv-test, train-val-test
    :param n_outerfolds: number of outerfolds relevant for nested-cv
    :param n_innerfolds: number of folds relevant for nested-cv and cv-test
    :param test_set_size_percentage: size of the test set relevant for cv-test and train-val-test
    :param val_set_size_percentage: size of the validation set relevant for train-val-test
    :param save_final_model: specify if the final model should be saved
    """

    # create Path
    results_directory_model = pathlib.Path(results_directory_model)
    if not check_functions.check_exist_directories(list_of_dirs=[results_directory_model]):
        raise Exception("See output above. Problems with specified directories")
    model_name = results_directory_model.parts[-1]
    if 'fromR' in model_name:
        import easypheno.model
        easypheno.model.__all__.extend(['_bayesfromR', 'bayesAfromR', 'bayesBfromR', 'bayesCfromR'])
    models = [model_name]
    data_dir = pathlib.Path(data_dir)
    # set save directory
    save_dir = data_dir if save_dir is None else pathlib.Path(save_dir)
    save_dir = save_dir if save_dir.is_absolute() else save_dir.resolve()
    # Check all arguments
    check_functions.check_all_specified_arguments(arguments=locals())
    # prepare all data files
    raw_data_functions.prepare_data_files(
        data_dir=data_dir, genotype_matrix_name=genotype_matrix, phenotype_matrix_name=phenotype_matrix,
        phenotype=phenotype, datasplit=datasplit, n_outerfolds=n_outerfolds, n_innerfolds=n_innerfolds,
        test_set_size_percentage=test_set_size_percentage, val_set_size_percentage=val_set_size_percentage,
        models=models, user_encoding=encoding, maf_percentage=maf_percentage
    )
    encoding = helper_functions.get_mapping_name_to_class()[model_name].standard_encoding \
        if encoding is None else encoding
    print("Load new dataset")
    new_dataset = base_dataset.Dataset(
        data_dir=data_dir, genotype_matrix_name=genotype_matrix, phenotype_matrix_name=phenotype_matrix,
        phenotype=phenotype, datasplit=datasplit, n_outerfolds=n_outerfolds, n_innerfolds=n_innerfolds,
        test_set_size_percentage=test_set_size_percentage, val_set_size_percentage=val_set_size_percentage,
        encoding=encoding, maf_percentage=maf_percentage
    )
    # train on new data
    if datasplit == 'train-val-test':
        datasplit_params = [val_set_size_percentage, test_set_size_percentage]
    elif datasplit == 'cv-test':
        datasplit_params = [n_innerfolds, test_set_size_percentage]
    elif datasplit == 'nested-cv':
        datasplit_params = [n_outerfolds, n_innerfolds]
    datasplit_subpath = helper_functions.get_subpath_for_datasplit(
        datasplit=datasplit, datasplit_params=datasplit_params
    )
    models_start_time = '+'.join(models) + '_' + datetime.datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
    base_path = save_dir.joinpath('results', genotype_matrix.split('.')[0], phenotype_matrix.split('.')[0],
                                  phenotype, datasplit + '_' + datasplit_subpath + '_MAF' + str(maf_percentage) +
                                  '_' + models_start_time)
    nested_offset = -1 if 'nested' in str(results_directory_model) else 0
    models_old = results_directory_model.parts[-2 + nested_offset].split('_')[3].split('+')
    results_file_path = results_directory_model.parents[0 - nested_offset].joinpath(
        'Results_overview_' + '_'.join(models_old) + '.csv')
    saved_outerfold_number = int(results_directory_model.parts[-2].split('_')[1]) if nested_offset != 0 else 0
    helper_functions.set_all_seeds()
    model_overview = {}
    overall_results = {}
    n_outerfolds = 1 if 'nested' not in datasplit else n_outerfolds
    for outerfold_number in range(n_outerfolds):
        print("train on new dataset")
        if 'nested' in datasplit:
            print("outerfold " + str(outerfold_number))
        model = _model_functions.retrain_model_with_results_file(
            results_file_path=results_file_path, model_name=model_name, datasplit=datasplit,
            outerfold_number=outerfold_number, dataset=new_dataset, saved_outerfold_number=saved_outerfold_number,
            saved_datasplit=results_directory_model.parts[-2 + nested_offset].split('_')[0]
        )
        if save_final_model:
            save_path = base_path.joinpath('outerfold_' + str(outerfold_number), model_name) \
                if 'nested' in datasplit else base_path.joinpath(model_name)
            if not save_path.exists():
                save_path.mkdir(parents=True)
            model.save_model(path=save_path, filename='final_retrained_model')
        outerfold_info = new_dataset.datasplit_indices['outerfold_' + str(outerfold_number)]
        X_test, y_test, sample_ids_test = \
            new_dataset.X_full[outerfold_info['test']], new_dataset.y_full[outerfold_info['test']], \
            new_dataset.sample_ids_full[outerfold_info['test']]
        X_retrain, y_retrain, sample_ids_retrain = \
            new_dataset.X_full[~np.isin(np.arange(len(new_dataset.X_full)), outerfold_info['test'])], \
            new_dataset.y_full[~np.isin(np.arange(len(new_dataset.y_full)), outerfold_info['test'])], \
            new_dataset.sample_ids_full[~np.isin(np.arange(len(new_dataset.sample_ids_full)), outerfold_info['test'])]
        y_pred_retrain = model.predict(X_in=X_retrain)
        y_pred_test = model.predict(X_in=X_test)

        # Evaluate and save results
        eval_scores = \
            eval_metrics.get_evaluation_report(y_pred=y_pred_test, y_true=y_test, task=model.task, prefix='test_')
        final_results = pd.DataFrame(index=range(0, new_dataset.y_full.shape[0]))
        final_results.at[0:len(sample_ids_retrain) - 1, 'sample_ids_retrain'] = sample_ids_retrain.flatten()
        final_results.at[0:len(y_pred_retrain) - 1, 'y_pred_retrain'] = y_pred_retrain.flatten()
        final_results.at[0:len(y_retrain) - 1, 'y_true_retrain'] = y_retrain.flatten()
        final_results.at[0:len(sample_ids_test) - 1, 'sample_ids_test'] = sample_ids_test.flatten()
        final_results.at[0:len(y_pred_test) - 1, 'y_pred_test'] = y_pred_test.flatten()
        final_results.at[0:len(y_test) - 1, 'y_true_test'] = y_test.flatten()

        for metric, value in eval_scores.items():
            final_results.at[0, metric] = value
        final_results.at[0, 'base_model_path'] = results_directory_model
        save_path = base_path.joinpath('outerfold_' + str(outerfold_number), model_name) \
            if 'nested' in datasplit else base_path.joinpath(model_name)
        if not save_path.exists():
            save_path.mkdir(parents=True)
        final_results.to_csv(save_path.joinpath('final_model_test_results.csv'),
                             sep=',', decimal='.', float_format='%.10f', index=False)
        key = 'outerfold_' + str(outerfold_number) if datasplit == 'nested-cv' else 'Test'
        best_params = model.optuna_trial.params \
            if issubclass(helper_functions.get_mapping_name_to_class()[model_name], _base_model.BaseModel) else None
        if issubclass(helper_functions.get_mapping_name_to_class()[model_name],
                      _torch_model.TorchModel) or \
                issubclass(helper_functions.get_mapping_name_to_class()[model_name],
                           _tensorflow_model.TensorflowModel):
            # additional attributes for torch and tensorflow models
            if 'n_epochs' not in best_params.keys():
                best_params['n_epochs'] = model.n_epochs
            if 'batch_size' not in best_params.keys():
                best_params['batch_size'] = model.batch_size
            best_params['early_stopping_point'] = model.early_stopping_point
        overall_results[key] = {'best_params': best_params, 'eval_metrics': eval_scores,
                                'runtime_metrics':
                                    {'process_time_mean': 0, 'process_time_std': 0,
                                     'process_time_max': 0, 'process_time_min': 0,
                                     'real_time_mean': 0, 'real_time_std': 0,
                                     'real_time_max': 0, 'real_time_min': 0}}
    model_overview[model_name] = overall_results
    nested_offset_new = -1 if 'nested' in datasplit else 0
    path_overview_file = \
        save_path.parents[0 - nested_offset_new].joinpath('Results_overview_' + '_'.join(models) + '.csv')
    helper_functions.save_model_overview_dict(model_overview=model_overview, save_path=path_overview_file)
    # generate feature importances
    feat_importance.post_generate_feature_importances(
        data_dir=str(data_dir),
        results_directory_genotype_level=str(save_dir.joinpath('results', genotype_matrix.split('.')[0]))
    )



import numpy as np
import pandas as pd

from ..preprocess import base_dataset
from . import helper_functions


def print_config_info(arguments: dict, dataset: base_dataset.Dataset, task: str):
    """
    Print info of current configuration.

    :param arguments: arguments specified by the user
    :param dataset: dataset used for optimization
    :param task: task that was detected
    """
    print('+++++++++++ CONFIG INFORMATION +++++++++++')
    print('Genotype Matrix: ' + arguments["genotype_matrix"])
    print('Phenotype Matrix: ' + arguments["phenotype_matrix"])
    print('Phenotype: ' + arguments["phenotype"])
    if arguments["encoding"] is not None:
        print('Encoding: ' + arguments["encoding"])
    if arguments["models"] != 'all':
        print('Models: ' + ', '.join(arguments["models"]))
    else:
        print('Models: ' + ', '.join(helper_functions.get_list_of_implemented_models()))
    print('Optuna Trials: ' + str(arguments["n_trials"]))
    if arguments["datasplit"] == 'train-val-test':
        datasplit_params = [arguments["val_set_size_percentage"], arguments["test_set_size_percentage"]]
    elif arguments["datasplit"] == 'cv-test':
        datasplit_params = [arguments["n_innerfolds"], arguments["test_set_size_percentage"]]
    elif arguments["datasplit"] == 'nested-cv':
        datasplit_params = [arguments["n_outerfolds"], arguments["n_innerfolds"]]
    print('Datasplit: ' + arguments["datasplit"] +
          ' (' + helper_functions.get_subpath_for_datasplit(datasplit=arguments["datasplit"],
                                                            datasplit_params=datasplit_params) + ')')
    print('MAF: ' + str(arguments["maf_percentage"]))
    print('Dataset Infos')
    print('- Task detected: ' + task)
    print('- No. of samples: ' + str(dataset.X_full.shape[0]) + ', No. of features: ' + str(dataset.X_full.shape[1]))
    print('- Encoding: ' + str(dataset.encoding))
    if task == 'classification':
        print('- Samples per class: ' + str(np.unique(dataset.y_full, return_counts=True)[1]))
    else:
        print('- Target variable statistics: ')
        print(pd.DataFrame(dataset.y_full).describe()[0])
    print('++++++++++++++++++++++++++++++++++++++++++++')



import argparse

from . import results_analysis

if __name__ == "__main__":
    """
    Run to generate the specified plot based on the results summary .csv file
    """

    parser = argparse.ArgumentParser()
    parser.add_argument("-rsp", "--results_summary_path", type=str,
                        help="Provide the full path to the results summary .csv file that you want to plot. "
                             "If not generated yet, run postprocess.results_analysis first.")
    parser.add_argument("-sd", "--save_dir", type=str, default=None,
                        help="Define save directory for the plots. Default is the same as results summary.")
    parser.add_argument("-plot", "--plot", type=str, default='heatmap',
                        help="select plot type: 'heatmap' | ")
    args = vars(parser.parse_args())
    results_summary_path = args['results_summary_path']
    save_dir = args['save_dir']
    if args['plot'] == 'heatmap':
        results_analysis.plot_heatmap_results(path_to_results_summary_csv=results_summary_path, save_dir=save_dir)
    else:
        print('Specified plot type ' + args['plot'] + ' not available. Please check help or source code.')



import sklearn

from . import _sklearn_model


class TemplateSklearnModel(_sklearn_model.SklearnModel):
    """
    Template file for a prediction model based on :obj:`~easypheno.model._sklearn_model.SklearnModel`

    See :obj:`~easypheno.model._base_model.BaseModel` for more information on the attributes.

    **Steps you have to do to add your own model:**

        1. Copy this template file and rename it according to your model (will be the name to call it later on on the command line)

        2. Rename the class and add it to *easypheno.model.__init__.py*

        3. Adjust the class attributes if necessary

        4. Define your model in *define_model()*

        5. Define the hyperparameters and ranges you want to use for optimization in *define_hyperparams_to_tune()*

        6. Test your new prediction model using toy data
    """
    standard_encoding = ...
    possible_encodings = [...]

    def define_model(self):
        """
        Definition of the actual prediction model.

        Use *param = self.suggest_hyperparam_to_optuna(PARAM_NAME_IN_DEFINE_HYPERPARAMS_TO_TUNE)* if you want to use
        the value of a hyperparameter that should be optimized.
        The function needs to return the model object.

        See :obj:`~easypheno.model._base_model.BaseModel` for more information.
        """
        ...

    def define_hyperparams_to_tune(self) -> dict:
        """
        Define the hyperparameters and ranges you want to optimize.
        Caution: they will only be optimized if you add them via *self.suggest_hyperparam_to_optuna(PARAM_NAME)* in *define_model()*

        See :obj:`~easypheno.model._base_model.BaseModel` for more information on the format and options.
        """
        return {
            'example_param_1': {
                'datatype': 'categorical',
                'list_of_values': ['cat', 'dog', 'elephant']
            },
            'example_param_2': {
                'datatype': 'float',
                'lower_bound': 0.05,
                'upper_bound': 0.95,
                'step': 0.05
            },
            'example_param_3': {
                'datatype': 'int',
                'lower_bound': 1,
                'upper_bound': 100
            }
        }
